import torch
from collections import OrderedDict

# https://gist.github.com/the-bass/0bf8aaa302f9ba0d26798b11e4dd73e3
def rename_state_dict_keys(source, key_transformation, target=None):
    """Rename state dict keys

    Args:
        source             -> Path to the saved state dict.
        key_transformation -> Function that accepts the old key names of the state
                            dict as the only argument and returns the new key name.
        target (optional)  -> Path at which the new state dict should be saved
                            (defaults to `source`)
        Example:
        Rename the key `layer.0.weight` `layer.1.weight` and keep the names of all
        other keys.

    Example:
        ```
        def key_transformation(old_key):
            if old_key == "layer.0.weight":
                return "layer.1.weight"
            return old_key
        rename_state_dict_keys(state_dict_path, key_transformation)
        ```
    """
    if target is None:
        target = source

    state_dict = torch.load(source)
    new_state_dict = OrderedDict()

    for key, value in state_dict.items():
        new_key = key_transformation(key)
        new_state_dict[new_key] = value

    torch.save(new_state_dict, target)



import os
import re
import glob


class ClassedCrawler:
    """ ClassedCrawler 

    Extracts objects of a specific set of classes only 
    """

    def __init__(
        self,
        data_folder="VeRi",
        train_folder="image_train",
        test_folder="image_test",
        query_folder="image_query",
        **kwargs
    ):
        self.metadata = {}

        self.data_folder = data_folder
        self.train_folder = os.path.join(self.data_folder, train_folder)
        self.test_folder = os.path.join(self.data_folder, test_folder)
        self.query_folder = os.path.join(self.data_folder, query_folder)

        self.logger = kwargs.get("logger")

        self.__verify(self.data_folder)
        self.__verify(self.train_folder)
        self.__verify(self.test_folder)
        self.__verify(self.query_folder)

        self.crawl()

    def __verify(self, folder):
        if not os.path.exists(folder):
            raise IOError(
                "Folder {data_folder} does not exist".format(data_folder=folder)
            )
        else:
            self.logger.info("Found {data_folder}".format(data_folder=folder))

    def crawl(self,):
        self.metadata["train"], self.metadata["test"], self.metadata["query"] = (
            {},
            {},
            {},
        )
        (
            self.metadata["train"]["crawl"],
            self.metadata["train"]["pids"],
            self.metadata["train"]["cids"],
            self.metadata["train"]["imgs"],
        ) = self.__crawl(self.train_folder, reset_labels=True)
        (
            self.metadata["test"]["crawl"],
            self.metadata["test"]["pids"],
            self.metadata["test"]["cids"],
            self.metadata["test"]["imgs"],
        ) = self.__crawl(self.test_folder)
        (
            self.metadata["query"]["crawl"],
            self.metadata["query"]["pids"],
            self.metadata["query"]["cids"],
            self.metadata["query"]["imgs"],
        ) = self.__crawl(self.query_folder)

        self.logger.info(
            "Train\tPIDS: {:6d}\tCIDS: {:6d}\tIMGS: {:8d}".format(
                self.metadata["train"]["pids"],
                self.metadata["train"]["cids"],
                self.metadata["train"]["imgs"],
            )
        )
        self.logger.info(
            "Test \tPIDS: {:6d}\tCIDS: {:6d}\tIMGS: {:8d}".format(
                self.metadata["test"]["pids"],
                self.metadata["test"]["cids"],
                self.metadata["test"]["imgs"],
            )
        )
        self.logger.info(
            "Query\tPIDS: {:6d}\tCIDS: {:6d}\tIMGS: {:8d}".format(
                self.metadata["query"]["pids"],
                self.metadata["query"]["cids"],
                self.metadata["query"]["imgs"],
            )
        )

    def __crawl(self, folder, reset_labels=False):
        imgs = glob.glob(os.path.join(folder, "*.jpg"))
        _re = re.compile(r"([\d]+)_[a-z]([\d]+)")
        pid_labeler = 0
        pid_tracker, cid_tracker = {}, {}
        crawler = []
        pid_counter, cid_counter, img_counter = 0, 0, 0
        for img in imgs:
            pid, cid = map(int, _re.search(img).groups())  # _re.search lol
            if pid < 0:
                continue  # ignore junk
            if cid < 0:
                continue  # ignore junk
            if pid not in pid_tracker:
                pid_tracker[pid] = pid_labeler if reset_labels else pid
                pid_labeler += 1
            if cid not in cid_tracker:
                cid_tracker[cid] = cid - 1
            crawler.append((img, pid_tracker[pid], cid - 1))  # cids start at 1 in data
        return crawler, len(pid_tracker), len(cid_tracker), len(crawler)



import json
from logging import Logger
import logging
import torch
import os
import shutil
from ednaml.config.EdnaMLConfig import EdnaMLConfig
from ednaml.crawlers import Crawler
import ednaml.loss.builders
from typing import Dict, List
from torch.utils.data import DataLoader
from ednaml.models.ModelAbstract import ModelAbstract

from ednaml.optimizer import BaseOptimizer
from ednaml.utils.LabelMetadata import LabelMetadata


class BaseTrainer:
    model: ModelAbstract
    loss_fn: Dict[str, ednaml.loss.builders.LossBuilder]  # output-name: lossBuilder
    optimizer: Dict[
        str, BaseOptimizer
    ]  # optimizer-name: optimizer             e.g. discriminator: params...
    loss_optimizer = Dict[str, List[BaseOptimizer]]  # output-name: optimizer
    scheduler: Dict[str, torch.optim.lr_scheduler._LRScheduler]  # TODO scheduler class
    loss_scheduler: Dict[
        str, List[torch.optim.lr_scheduler._LRScheduler]
    ]  # output-name: scheduler

    skipeval: bool
    train_loader: DataLoader
    test_loader: DataLoader

    epochs: int
    logger: Logger
    global_batch: int
    global_epoch: int
    num_losses: int
    losses: Dict[str, List[int]]  # output-name: losses
    # config metadata, trainer metadata ???????
    metadata: Dict[str, str]
    labelMetadata: LabelMetadata
    logger: logging.Logger

    def __init__(
        self,
        model: ModelAbstract,
        loss_fn: List[ednaml.loss.builders.LossBuilder],
        optimizer: torch.optim.Optimizer,
        loss_optimizer: List[torch.optim.Optimizer],
        scheduler: torch.optim.lr_scheduler._LRScheduler,
        loss_scheduler: torch.optim.lr_scheduler._LRScheduler,
        train_loader: DataLoader,
        test_loader: DataLoader,
        epochs: int,
        skipeval: bool,
        logger: Logger,
        crawler: Crawler,
        config: EdnaMLConfig,
        labels: LabelMetadata,
        **kwargs
    ):

        self.model = model
        self.loss_fn_order = {
            idx: lossbuilder.loss_labelname for idx, lossbuilder in enumerate(loss_fn)
        }
        self.loss_fn = {
            lossbuilder.loss_labelname: lossbuilder for lossbuilder in loss_fn
        }
        self.num_losses = len(self.loss_fn)
        self.losses = {lossname: [] for lossname in self.loss_fn}
        self.loss_optimizer = {
            self.loss_fn_order[idx]: loss_optimizer_content
            for idx, loss_optimizer_content in enumerate(loss_optimizer)
        }
        if type(loss_scheduler) is list:
            self.loss_scheduler = {
                self.loss_fn_order[idx]: loss_scheduler_content
                for idx, loss_scheduler_content in enumerate(loss_scheduler)
            }
        else:
            self.loss_scheduler = {
                self.loss_fn_order[idx]: loss_scheduler
                for idx in range(self.num_losses)
            }

        self.optimizer = optimizer
        self.scheduler = scheduler
        self.skipeval = skipeval
        self.train_loader = train_loader
        self.test_loader = test_loader

        self.epochs = epochs
        self.logger = logger

        self.global_batch = 0  # Current batch number in the epoch
        self.global_epoch = 0

        self.metadata = {}
        self.labelMetadata = labels
        self.crawler = crawler
        self.config = config

        self.buildMetadata(
            crawler=crawler.classes, config=json.loads(config.export("json"))
        )

    def buildMetadata(self, **kwargs):
        for keys in kwargs:
            self.metadata[keys] = kwargs.get(keys)

    def setup(
        self,
        step_verbose: int = 5,
        save_frequency: int = 5,
        test_frequency: int = 5,
        save_directory: str = "./checkpoint/",
        save_backup: bool = False,
        backup_directory: str = None,
        gpus: int = 1,
        fp16: bool = False,
        model_save_name: str = None,
        logger_file: str = None,
    ):
        self.step_verbose = step_verbose
        self.save_frequency = save_frequency
        self.test_frequency = test_frequency
        self.save_directory = save_directory
        self.backup_directory = None
        self.model_save_name = model_save_name
        self.logger_file = logger_file
        self.save_backup = save_backup
        if self.save_backup:
            self.backup_directory = backup_directory
            os.makedirs(self.backup_directory, exist_ok=True)
        os.makedirs(self.save_directory, exist_ok=True)
        self.saveMetadata()

        self.gpus = gpus

        if self.gpus != 1:
            raise NotImplementedError()

        self.model.cuda()

        self.fp16 = fp16
        # if self.fp16 and self.apex is not None:
        #    self.model, self.optimizer = self.apex.amp.initialize(self.model, self.optimizer, opt_level='O1')

    def saveMetadata(self):
        print("NOT saving metadata. saveMetadata() function not set up.")

    def save(self):
        self.logger.info("Saving model, optimizer, and scheduler.")
        MODEL_SAVE = self.model_save_name + "_epoch%i" % self.global_epoch + ".pth"
        TRAINING_SAVE = (
            self.model_save_name + "_epoch%i" % self.global_epoch + "_training.pth"
        )

        save_dict = {}
        save_dict["optimizer"] = self.optimizer.state_dict()
        save_dict["scheduler"] = self.scheduler.state_dict()
        save_dict["loss_fn"] = {
            lossname: self.loss_fn[lossname].state_dict() for lossname in self.loss_fn
        }
        save_dict["loss_optimizer"] = {
            lossname: (
                self.loss_optimizer[lossname].state_dict()
                if self.loss_optimizer[lossname] is not None
                else None
            )
            for lossname in self.loss_optimizer
        }
        save_dict["loss_scheduler"] = {
            lossname: (
                self.loss_scheduler[lossname].state_dict()
                if self.loss_scheduler[lossname] is not None
                else None
            )
            for lossname in self.loss_scheduler
        }

        # save_dict["loss_optimizer"] = [self.loss_optimizer[idx].state_dict() if self.loss_optimizer[idx] is not None else None for idx in range(self.num_losses)]
        # save_dict["loss_scheduler"] = [self.loss_scheduler[idx].state_dict() if self.loss_scheduler[idx] is not None else None for idx in range(self.num_losses)]

        torch.save(
            self.model.state_dict(), os.path.join(self.save_directory, MODEL_SAVE)
        )
        torch.save(save_dict, os.path.join(self.save_directory, TRAINING_SAVE))

        if self.save_backup:
            shutil.copy2(
                os.path.join(self.save_directory, MODEL_SAVE), self.backup_directory
            )
            shutil.copy2(
                os.path.join(self.save_directory, TRAINING_SAVE), self.backup_directory
            )

            self.logger.info(
                "Performing drive backup of model, optimizer, and scheduler."
            )

            LOGGER_SAVE = os.path.join(self.backup_directory, self.logger_file)
            if os.path.exists(LOGGER_SAVE):
                os.remove(LOGGER_SAVE)
            shutil.copy2(
                os.path.join(self.save_directory, self.logger_file), LOGGER_SAVE
            )

    def load(self, load_epoch):
        self.logger.info(
            "Resuming training from epoch %i. Loading saved state from %i"
            % (load_epoch + 1, load_epoch)
        )
        model_load = self.model_save_name + "_epoch%i" % load_epoch + ".pth"
        training_load = self.model_save_name + "_epoch%i" % load_epoch + "_training.pth"

        if self.save_backup:
            self.logger.info(
                "Loading model, optimizer, and scheduler from drive backup."
            )
            model_load_path = os.path.join(self.backup_directory, model_load)
            training_load_path = os.path.join(self.backup_directory, training_load)

        else:
            self.logger.info(
                "Loading model, optimizer, and scheduler from local backup."
            )
            model_load_path = os.path.join(self.save_directory, model_load)
            training_load_path = os.path.join(self.save_directory, training_load)

        self.model.load_state_dict(torch.load(model_load_path))
        self.logger.info("Finished loading model state_dict from %s" % model_load_path)

        checkpoint = torch.load(training_load_path)
        self.optimizer.load_state_dict(checkpoint["optimizer"])
        self.logger.info(
            "Finished loading optimizer state_dict from %s" % training_load_path
        )
        self.scheduler.load_state_dict(checkpoint["scheduler"])
        self.logger.info(
            "Finished loading scheduler state_dict from %s" % training_load_path
        )

        for lossname in self.loss_fn:
            self.loss_fn[lossname].load_state_dict(checkpoint["loss_fn"][lossname])
            if self.loss_optimizer[lossname] is not None:
                self.loss_optimizer[lossname].load_state_dict(
                    checkpoint["loss_optimizer"][lossname]
                )
            if self.loss_scheduler[lossname] is not None:
                self.loss_scheduler[lossname].load_state_dict(
                    checkpoint["loss_scheduler"][lossname]
                )

            self.logger.info(
                "Finished loading loss state_dict from %s" % training_load_path
            )

        # for idx in range(self.num_losses):
        #    if self.loss_optimizer[idx] is not None:
        #        self.loss_optimizer[idx].load_state_dict(checkpoint["loss_optimizer"][idx])
        #    if self.loss_scheduler[idx] is not None:
        #        self.loss_scheduler[idx].load_state_dict(checkpoint["loss_scheduler"][idx])

    def train(self, continue_epoch=0):
        self.logger.info("Starting training")
        self.logger.info("Logging to:\t%s" % self.logger_file)
        self.logger.info(
            "Models will be saved to local directory:\t%s" % self.save_directory
        )
        if self.save_backup:
            self.logger.info(
                "Models will be backed up to drive directory:\t%s"
                % self.backup_directory
            )
        self.logger.info(
            "Models will be saved with base name:\t%s_epoch[].pth"
            % self.model_save_name
        )
        self.logger.info(
            "Optimizers will be saved with base name:\t%s_epoch[]_optimizer.pth"
            % self.model_save_name
        )
        self.logger.info(
            "Schedulers will be saved with base name:\t%s_epoch[]_scheduler.pth"
            % self.model_save_name
        )

        if continue_epoch > 0:
            load_epoch = continue_epoch - 1
            self.load(load_epoch)

        if not self.skipeval:
            self.logger.info("Performing initial evaluation...")
            self.initial_evaluate()
        else:
            self.logger.info("Skipping initial evaluation.")

        self.logger.info("Starting training from %i" % continue_epoch)
        for epoch in range(self.epochs):
            if epoch >= continue_epoch:
                self.epoch_step(epoch)
            else:
                self.global_epoch = epoch + 1

    def initial_evaluate(self):
        """Evaluation of model before we start training
        """
        self.evaluate()

    def epoch_step(self, epoch):
        """Trains model for an epoch.
        """
        raise NotImplementedError()

    def evaluate(self):
        logit_labels, true_labels, features = self.evaluate_impl()
        return logit_labels, true_labels, self.crawler.classes, features

    def evaluate_impl(self):
        raise NotImplementedError



from torch import nn


class Loss(nn.Module):
    def __init__(self, lossname=None, metadata=None, **kwargs):
        super().__init__()

    def forward(self):
        raise NotImplementedError()


from ednaml.loss.CenterLoss import CenterLoss
from ednaml.loss.ClusterLoss import ClusterLoss
from ednaml.loss.CompactContrastiveLoss import CompactContrastiveLoss
from ednaml.loss.ContrastiveLoss import ContrastiveLoss
from ednaml.loss.MarginLoss import MarginLoss
from ednaml.loss.ProxyNCA import ProxyNCA
from ednaml.loss.SoftmaxLabelSmooth import SoftmaxLabelSmooth
from ednaml.loss.SoftmaxLogitsLoss import SoftmaxLogitsLoss
from ednaml.loss.TripletLoss import TripletLoss



import torch
from torch import nn
from ednaml.loss import Loss
from ednaml.utils.math import pairwise_distance


class ProxyNCA(Loss):
    """ProxyNCA Loss

    Performs softmax with label smoothing.

    Args (kwargs only):
        classes (int): Number of classes in training set
        embedding_size (int): Number of embedding features
        smoothing (float, 0.3): Smoothing constant. Default 0.0
        normalization (float, 3.0): Normalization constant. Default 0.0

    Methods: 
        __call__: Returns loss given features and labels.

    """

    def __init__(self, **kwargs):
        super(ProxyNCA, self).__init__()

        self.classes = kwargs.get("classes")
        self.embedding = kwargs.get("embedding_size")
        self.DIV_CONST = 8
        self.SMOOTHING = kwargs.get("smoothing", 0.1)
        self.NORMALIZATION = kwargs.get("normalization", 3.0)
        self.logsoftmax = nn.LogSoftmax(dim=-1)
        self.proxies = nn.Parameter(
            torch.randn(self.classes, self.embedding) / 8
        ).cuda()

    def forward(self, features, labels):
        normalized_proxy = self.NORMALIZATION * torch.nn.functional.normalize(
            self.proxies, p=2, dim=-1
        )
        normalized_logits = self.NORMALIZATION * torch.nn.functional.normalize(
            features, p=2, dim=-1
        )

        dist = pairwise_distance(
            torch.cat([normalized_logits, normalized_proxy]), squared=True
        )
        # the pairwise_distance is for all logits and proxies to each other. We just need logits to proxies, so we take the bottom left quadrant.
        dist = dist[: features.size()[0], features.size()[0] :]
        log_probs = self.logsoftmax(dist)

        # smooth labels
        labels = torch.zeros(log_probs.size()).scatter_(
            1, labels.unsqueeze(1).data.cpu(), 1
        )
        labels = labels.cuda()
        labels = (
            1 - (self.SMOOTHING + (self.SMOOTHING / self.embedding))
        ) * labels + self.SMOOTHING / self.embedding

        # cross entropy with distances as logits, one hot labels
        # note that compared to proxy nca, positive not excluded in denominator
        loss = (-labels * log_probs).mean(0).sum()
        return loss



import importlib
import os, shutil, logging, glob, re, json
from typing import Dict, List, Type
import warnings
import kaptan
from torchinfo import ModelStatistics
from ednaml.config.EdnaMLConfig import EdnaMLConfig
from ednaml.crawlers import Crawler
from ednaml.datareaders import DataReader
from ednaml.generators import ImageGenerator
from ednaml.loss.builders import LossBuilder
from ednaml.models.ModelAbstract import ModelAbstract
from ednaml.optimizer import BaseOptimizer
from ednaml.optimizer.StandardLossOptimizer import StandardLossOptimizer
from ednaml.loss.builders import ClassificationLossBuilder
from ednaml.trainer.BaseTrainer import BaseTrainer
import ednaml.utils
import torch
from torchinfo import summary
from ednaml.utils.LabelMetadata import LabelMetadata
import ednaml.utils.web
from ednaml.utils.SaveMetadata import SaveMetadata
import logging

"""TODO

Verbosity = 0 -> create no logger and use a dummy that print nothing anywhere
"""


class EdnaML:
    logLevels = {0: logging.NOTSET, 1: logging.ERROR, 2: logging.INFO, 3: logging.DEBUG}
    labelMetadata: LabelMetadata
    modelStatistics: ModelStatistics
    model: ModelAbstract
    loss_function_array: List[LossBuilder]
    loss_optimizer_array: List[torch.optim.Optimizer]
    optimizer: List[torch.optim.Optimizer]
    scheduler: List[torch.optim.lr_scheduler._LRScheduler]
    loss_scheduler: List[torch.optim.lr_scheduler._LRScheduler]
    previous_stop: int
    trainer: BaseTrainer
    crawler: Crawler
    train_generator: ImageGenerator
    test_generator: ImageGenerator
    cfg: EdnaMLConfig

    def __init__(
        self,
        config: str = "config.yaml",
        mode: str = "train",
        weights: str = None,
        logger: logging.Logger = None,
        verbose: int = 2,
    ):
        """Initializes the EdnaML object with the associated config, mode, weights, and verbosity. 
        Sets up the logger, as well as local logger save directory. 

        Args:
            config (str, optional): Path to the Edna config file. Defaults to "config.yaml".
            mode (str, optional): Either `train` or `test`. Defaults to "train".
            weights (str, optional): The path to the weights file. Defaults to None.
            logger (logging.Logger, optional): A logger. If no logger is provided, EdnaML will construct its own. Defaults to None.
            verbose (int, optional): Logging verbosity. Defaults to 2.
        """

        self.config = config
        self.mode = mode
        self.weights = weights
        self.pretrained_weights = None
        self.verbose = verbose

        self.cfg = self.buildConfig(config)
        self.saveMetadata = self.buildSaveMetadata()

        self.gpus = torch.cuda.device_count()
        self.drive_backup = self.cfg.SAVE.DRIVE_BACKUP
        self.previous_stop = 0
        self.epochs = self.cfg.EXECUTION.EPOCHS
        self.skipeval = self.cfg.EXECUTION.SKIPEVAL

        self.step_verbose = self.cfg.LOGGING.STEP_VERBOSE
        self.save_frequency = self.cfg.SAVE.SAVE_FREQUENCY
        self.test_frequency = self.cfg.EXECUTION.TEST_FREQUENCY
        self.fp16 = self.cfg.OPTIMIZER.FP16

        self.logger = self.buildLogger(logger=logger)

        self.makeSaveDirectories()

    def buildConfig(self, config: str, handler="yaml") -> EdnaMLConfig:
        """Builds the internal kaptan configuration object

        Args:
            config (str): Path to config file
            handler (str, optional): Handler for kaptan. Defaults to "yaml".
        """
        return EdnaMLConfig(config)
        # kaptan.Kaptan(handler=handler)
        # self.cfg = self.cfg.import_config(config)

    def setup(self):
        """setup() completes initial setup, allowing you to proceed with the core ml pipeline
        of dataloading, model building, etc
        """
        self.printConfiguration()
        self.downloadModelWeights()

    def downloadModelWeights(self):
        """Downloads model weights specified in the configuration if `weights` were not passed into EdnaML and if model weights are supported.

        TODO -- do not throw error for no weights or missing base, if this is a new architecture to be trained from scratch
        Raises:
            Warning: If there are no pre-downloaded weights, and the model architecture is unsupported
        """
        if self.weights is not None:
            self.logger.info("Not downloading weights. Weights path already provided.")

        if self.mode == "train":
            self._download_weights_from_base(self.cfg.MODEL.MODEL_BASE)
        else:
            if self.weights is None:
                warnings.warn(
                    "Mode is `test` but weights is `None`. This will cause issues when EdnaML attempts to load weights"
                )

    def _download_weights_from_base(self, model_base: str):
        """Downloads weights from a model_base parameter directly from pytorch's servers.

        Args:
            model_base (str): A supported `model_base`, e.g. resnet18, resnet34. See `utils.web.model_weights`.
        """
        from utils import model_weights

        if model_base in model_weights:
            if os.path.exists(model_weights[model_base][1]):
                pass
            else:
                self.logger.info(
                    "Model weights file {} does not exist. Downloading.".format(
                        model_weights[model_base][1]
                    )
                )
                ednaml.utils.web.download(
                    model_weights[model_base][1], model_weights[model_base][0]
                )
            self.pretrained_weights = model_weights[model_base][1]
        else:
            warnings.warn(
                "Model %s is not available. Please choose one of the following: %s if you want to load pretrained weights"
                % (model_base, str(model_weights.keys()))
            )

    def quickSetup(self):
        """Performs a `quickstart` set-up of EdnaML
        """
        self.setup()
        self.getPreviousStop()  # TODO -- load weights for previous stop outside of trainer...

        self.buildDataloaders()

        self.buildModel()
        self.loadWeights()
        self.getModelSummary()
        self.buildOptimizer()
        self.buildScheduler()

        self.buildLossArray()
        self.buildLossOptimizer()
        self.buildLossScheduler()

        self.buildTrainer()

    def train(self):
        self.trainer.train(continue_epoch=self.previous_stop)

    def eval(self):
        return self.trainer.evaluate()

    def buildTrainer(self):
        """Builds the EdnaML trainer and sets it up
        """
        ExecutionTrainer: Type[BaseTrainer] = importlib.import_module(
            self.cfg.EXECUTION, package="ednaml.trainer"
        )
        self.logger.info(
            "Loaded {} from {} to build Trainer".format(
                self.cfg.EXECUTION.TRAINER, "trainer"
            )
        )

        self.trainer = ExecutionTrainer(
            model=self.model,
            loss_fn=self.loss_function_array,
            optimizer=self.optimizer,
            loss_optimizer=self.loss_optimizer_array,
            scheduler=self.scheduler,
            loss_scheduler=self.loss_scheduler,
            train_loader=self.train_generator.dataloader,
            test_loader=self.test_generator.dataloader,
            epochs=self.epochs,
            skipeval=self.skipeval,
            logger=self.logger,
            crawler=self.crawler,
            config=self.cfg,
            labels=self.labelMetadata,
        )
        self.trainer.setup(
            step_verbose=self.step_verbose,
            save_frequency=self.save_frequency,
            test_frequency=self.test_frequency,
            save_directory=self.saveMetadata.MODEL_SAVE_FOLDER,
            save_backup=self.drive_backup,
            backup_directory=self.saveMetadata.CHECKPOINT_DIRECTORY,
            gpus=self.gpus,
            fp16=self.fp16,
            model_save_name=self.saveMetadata.MODEL_SAVE_NAME,
            logger_file=self.saveMetadata.LOGGER_SAVE_NAME,
        )

    def getPreviousStop(self):
        """Gets the previous stop, if any, of the trainable model by checking local save directory, as well as a network directory. 
        """
        if self.drive_backup:
            fl_list = glob.glob(
                os.path.join(self.saveMetadata.CHECKPOINT_DIRECTORY, "*.pth")
            )
        else:
            fl_list = glob.glob(
                os.path.join(self.saveMetadata.MODEL_SAVE_FOLDER, "*.pth")
            )
        _re = re.compile(r".*epoch([0-9]+)\.pth")
        previous_stop = [
            int(item[1])
            for item in [_re.search(item) for item in fl_list]
            if item is not None
        ]
        if len(previous_stop) == 0:
            self.previous_stop = 0
            self.logger.info("No previous stop detected. Will start from epoch 0")
        else:
            self.previous_stop = max(previous_stop) + 1
            self.logger.info(
                "Previous stop detected. Will attempt to resume from epoch %i"
                % self.previous_stop
            )

    def buildOptimizer(self):
        """Builds the optimizer for the model
        """
        optimizer_builder: Type[BaseOptimizer] = importlib.import_module(
            self.cfg.EXECUTION.OPTIMIZER_BUILDER, package="ednaml.optimizer"
        )
        self.logger.info(
            "Loaded {} from {} to build Optimizer model".format(
                self.cfg.EXECUTION.OPTIMIZER_BUILDER, "ednaml.optimizer"
            )
        )

        # Optimizers are in a list...
        OPT_array = [
            optimizer_builder(
                name=optimizer_item.OPTIMIZER_NAME,
                optimizer=optimizer_item.OPTIMIZER,
                base_lr=optimizer_item.BASE_LR,
                lr_bias=optimizer_item.LR_BIAS_FACTOR,
                weight_decay=optimizer_item.WEIGHT_DECAY,
                weight_bias=optimizer_item.WEIGHT_BIAS_FACTOR,
                opt_kwargs=optimizer_item.OPTIMIZER_KWARGS,
                gpus=self.gpus,
            )
            for optimizer_item in self.cfg.OPTIMIZER
        ]
        self.optimizer = [
            OPT.build(self.model) for OPT in OPT_array
        ]  # TODO deal with singleton vs multiple optimizers...
        self.logger.info("Built optimizer")

    def buildScheduler(self):
        """Builds the scheduler for the model
        """
        self.scheduler = [None] * len(self.cfg.SCHEDULER)
        for idx, scheduler_item in enumerate(self.cfg.SCHEDULER):
            try:  # We first check if scheduler is part of torch's provided schedulers.
                scheduler = importlib.import_module(
                    scheduler_item.LR_SCHEDULER, package="torch.optim.lr_scheduler"
                )
            except (
                ModuleNotFoundError,
                AttributeError,
            ):  # If it fails, then we try to import from schedulers implemented in scheduler/ folder
                scheduler = importlib.import_module(
                    scheduler_item.LR_SCHEDULER, package="ednaml.scheduler"
                )
            self.scheduler[idx] = scheduler(
                self.optimizer[idx], last_epoch=-1, **scheduler_item.LR_KWARGS
            )
        self.logger.info("Built scheduler")

    def buildLossArray(self):
        """Builds the loss function array using the LOSS list in the provided configuration
        """
        self.loss_function_array = [
            ClassificationLossBuilder(
                loss_functions=loss_item.LOSSES,
                loss_lambda=loss_item.LAMBDAS,
                loss_kwargs=loss_item.KWARGS,
                name=loss_item.NAME,  # get("NAME", None),
                label=loss_item.LABEL,  # get("LABEL", None),
                metadata=self.labelMetadata,
                **{"logger": self.logger}
            )
            for loss_item in self.cfg.LOSS
        ]
        self.logger.info("Built loss function")

    def buildLossOptimizer(self):
        """Builds the Optimizer for loss functions, if the loss functions have learnable parameters (e.g. proxyNCA loss)

        self.loss_function_array contains a list of LossBuilders. Each LossBuilder 
        is for a specific output. Here, we build an array of StandardLossOptimizers,
        one StandardLossOptimizer for each LossBuilder. Each StandardLossOptimizer 
        takes as input the same arguments as an Optimizer. However, the name parameter
        should be the name of the LossBuilder it is targeting.

        If there is no LOSS_OPTIMIZER section in the configuration, the EdnaML config creates a default
        LOSS_OPTIMIZER, whose parameters we will use.

        """
        loss_optimizer_name_dict = {
            loss_optim_item.OPTIMIZER_NAME: loss_optim_item
            for loss_optim_item in self.cfg.LOSS_OPTIMIZER
        }
        initial_key = list(loss_optimizer_name_dict.keys())[0]
        LOSS_OPT: List[StandardLossOptimizer] = [None] * len(self.loss_function_array)
        for idx, loss_builder in enumerate(self.loss_function_array):
            if loss_builder.loss_labelname in loss_optimizer_name_dict:
                # Means we have an optimizer corresponding to this loss
                lookup_key = loss_builder.loss_labelname
            else:  # We will use the first optimizer (either default or otherwise, etc, for this)
                lookup_key = initial_key
            LOSS_OPT[idx] = StandardLossOptimizer(
                name=loss_optimizer_name_dict[lookup_key].OPTIMIZER_NAME,
                optimizer=loss_optimizer_name_dict[lookup_key].OPTIMIZER,
                base_lr=loss_optimizer_name_dict[lookup_key].BASE_LR,
                lr_bias=loss_optimizer_name_dict[lookup_key].LR_BIAS_FACTOR,
                gpus=self.gpus,
                weight_bias=loss_optimizer_name_dict[lookup_key].WEIGHT_BIAS_FACTOR,
                weight_decay=loss_optimizer_name_dict[lookup_key].WEIGHT_DECAY,
                opt_kwargs=loss_optimizer_name_dict[lookup_key].OPTIMIZER_KWARGS,
            )

        # Note: build returns None if there are no differentiable parameters
        self.loss_optimizer_array = [
            loss_opt.build(loss_builder=self.loss_function_array[idx])
            for idx, loss_opt in enumerate(LOSS_OPT)
        ]
        self.logger.info("Built loss optimizer")

    def buildLossScheduler(self):
        """Builds the scheduler for the loss functions, if the functions have learnable parameters and corresponding optimizer.
        """
        loss_scheduler_name_dict = {
            loss_schedule_item.SCHEDULER_NAME: loss_schedule_item
            for loss_schedule_item in self.cfg.LOSS_SCHEDULER
        }
        initial_key = list(loss_scheduler_name_dict.keys())[0]
        self.loss_scheduler = [None] * len(self.loss_optimizer_array)

        for idx, loss_optimizer in enumerate(self.loss_optimizer_array):
            if (
                loss_optimizer is not None
            ):  # In case loss has differentiable parameters, so the optimizer is not None...we look for the loss name
                if (
                    self.loss_function_array[idx].loss_labelname
                    in loss_scheduler_name_dict
                ):
                    lookup_key = self.loss_function_array[idx].loss_labelname
                else:
                    lookup_key = initial_key

                try:  # We first check if scheduler is part of torch's provided schedulers.
                    loss_scheduler = importlib.import_module(
                        loss_scheduler_name_dict[lookup_key].LR_SCHEDULER,
                        package="torch.optim.lr_scheduler",
                    )
                except (
                    ModuleNotFoundError,
                    AttributeError,
                ):  # If it fails, then we try to import from schedulers implemented in scheduler/ folder
                    loss_scheduler = importlib.import_module(
                        loss_scheduler_name_dict[lookup_key].LR_SCHEDULER,
                        package="ednaml.scheduler",
                    )
                self.loss_scheduler[idx] = loss_scheduler(
                    loss_optimizer,
                    last_epoch=-1,
                    **loss_scheduler_name_dict[lookup_key].LR_KWARGS
                )
                self.logger.info("Built loss scheduler")

    def _setModelTestMode(self):
        """Sets model to test mode if EdnaML is in testing mode
        """
        if self.mode == "test":
            self.model.eval()

    def _setModelTrainMode(self):
        """Sets the model to train mode if EdnaML is in training mode
        """
        if self.mode == "train":
            self.model.train()

    def _covert_model_kwargs(self) -> Dict[str, int]:
        """Converts the model_kwargs inside config into the correct format, depending on whether it is provided directly in yaml format, or as a json string

        Returns:
            Dict[str,Union[str,int]]: Corrected model_kwargs dictionary
        """

        if (
            type(self.cfg.MODEL.MODEL_KWARGS) is dict
        ):  # Compatibility with old configs. TODO fix all old configs.
            model_kwargs_dict = self.cfg.MODEL.MODEL_KWARGS
        elif type(self.cfg.MODEL.MODEL_KWARGS) is None:
            model_kwargs_dict = {}
        elif type(self.cfg.MODEL.MODEL_KWARGS) is str:
            raise ValueError("Outdated model_kwargs as str")
            # model_kwargs_dict = json.loads(self.cfg.MODEL.MODEL_KWARGS)
        return model_kwargs_dict

    def buildModel(self):
        """Builds an EdnaML model using the configuration. If there are pretrained weights, they are provided through the config to initialize the model.
        """
        model_builder = importlib.import_module(
            self.cfg.MODEL.BUILDER, package="ednaml.models"
        )
        self.logger.info(
            "Loaded {} from {} to build model".format(self.cfg.MODEL.BUILDER, "models")
        )

        # model_kwargs = self._covert_model_kwargs()

        # TODO!!!!!!!
        self.model: ModelAbstract = model_builder(
            arch=self.cfg.MODEL.MODEL_ARCH,
            base=self.cfg.MODEL.MODEL_BASE,
            weights=self.pretrained_weights,
            metadata=self.labelMetadata,
            normalization=self.cfg.MODEL.MODEL_NORMALIZATION,
            **self.cfg.MODEL.MODEL_KWARGS
        )
        self.logger.info(
            "Finished instantiating model with {} architecture".format(
                self.cfg.MODEL.MODEL_ARCH
            )
        )

    def loadWeights(self):
        """If in `test` mode, load weights from weights path. Otherwise, partially load what is possible from given weights path, if given.
        Note that for training mode, weights are downloaded from pytoch to be loaded if pretrained weights are desired.
        """
        if self.mode == "test":
            self.model.load_state_dict(torch.load(self.weights))
        else:
            if self.weights is None:
                self.logger.info("No saved model weights provided.")
            else:
                if (
                    self.weights != ""
                ):  # Load weights if train and starting from a another model base...
                    self.logger.info(
                        "Commencing partial model load from {}".format(self.weights)
                    )
                    self.model.partial_load(self.weights)
                    self.logger.info(
                        "Completed partial model load from {}".format(self.weights)
                    )

    def getModelSummary(self):
        """Gets the model summary using `torchinfo` and saves it as a ModelStatistics object
        """
        self.model.cuda()
        self.model_summary = summary(
            self.model,
            input_size=(
                self.cfg.TRANSFORMATION.BATCH_SIZE,
                self.cfg.TRANSFORMATION.CHANNELS,
                *self.cfg.TRANSFORMATION.SHAPE,
            ),
            col_names=[
                "input_size",
                "output_size",
                "num_params",
                "kernel_size",
                "mult_adds",
            ],
            depth=3,
            mode="train",
            verbose=1,
        )
        self.logger.info(str(self.model_summary))

    def buildDataloaders(self):
        """Sets up the datareader classes and builds the train and test dataloaders
        """
        data_reader: Type[DataReader] = importlib.import_module(
            self.cfg.EXECUTION.DATAREADER.DATAREADER, package="ednaml.datareaders"
        )
        data_reader_instance = data_reader()
        # data_crawler is now data_reader.CRAWLER
        self.logger.info("Reading data with DataReader %s" % data_reader.name)
        # Update the generator...if needed
        if self.cfg.EXECUTION.DATAREADER.GENERATOR is not None:
            data_reader_instance.GENERATOR = importlib.import_module(
                self.cfg.EXECUTION.DATAREADER.GENERATOR, package="ednaml.generators"
            )

        self.crawler = self.buildCrawlerInstance(data_reader=data_reader_instance)

        self.buildTrainDataloader(data_reader_instance, self.crawler)
        self.buildTestDataloader(data_reader_instance, self.crawler)

    def buildCrawlerInstance(self, data_reader: DataReader) -> Crawler:
        """Builds a Crawler instance from the data_reader's provided crawler class in `data_reader.CRAWLER`

        Args:
            data_reader (DataReader): A DataReader class

        Returns:
            Crawler: A Crawler instanece for this experiment
        """
        return data_reader.CRAWLER(
            logger=self.logger, **self.cfg.EXECUTION.DATAREADER.CRAWLER_ARGS
        )

    def buildTrainDataloader(self, data_reader: DataReader, crawler_instance: Crawler):
        """Builds a train dataloader instance given the data_reader class and a crawler instance that has been initialized

        Args:
            data_reader (DataReader): A datareader class
            crawler_instance (Crawler): A crawler instance
        """
        self.train_generator: ImageGenerator = data_reader.GENERATOR(
            gpus=self.gpus,
            i_shape=self.cfg.TRANSFORMATION.SHAPE,
            normalization_mean=self.cfg.TRANSFORMATION.NORMALIZATION_MEAN,
            normalization_std=self.cfg.TRANSFORMATION.NORMALIZATION_STD,
            normalization_scale=1.0 / self.cfg.TRANSFORMATION.NORMALIZATION_SCALE,
            h_flip=self.cfg.TRANSFORMATION.H_FLIP,
            t_crop=self.cfg.TRANSFORMATION.T_CROP,
            rea=self.cfg.TRANSFORMATION.RANDOM_ERASE,
            rea_value=self.cfg.TRANSFORMATION.RANDOM_ERASE_VALUE,
            **self.cfg.EXECUTION.DATAREADER.GENERATOR_ARGS
        )

        self.train_generator.setup(
            crawler_instance,
            mode="train",
            batch_size=self.cfg.TRANSFORMATION.BATCH_SIZE,
            workers=self.cfg.TRANSFORMATION.WORKERS,
            **self.cfg.EXECUTION.DATAREADER.DATASET_ARGS
        )
        self.logger.info("Generated training data generator")
        self.labelMetadata = self.train_generator.num_entities
        self.logger.info(
            "Running classification model with classes: %s"
            % str(self.labelMetadata.metadata)
        )

    def buildTestDataloader(self, data_reader: DataReader, crawler_instance: Crawler):
        """Builds a test dataloader instance given the data_reader class and a crawler instance that has been initialized

        Args:
            data_reader (DataReader): A datareader class
            crawler_instance (Crawler): A crawler instance
        """
        self.test_generator: ImageGenerator = data_reader.GENERATOR(
            gpus=self.gpus,
            i_shape=self.cfg.TRANSFORMATION.SHAPE,
            normalization_mean=self.cfg.TRANSFORMATION.NORMALIZATION_MEAN,
            normalization_std=self.cfg.TRANSFORMATION.NORMALIZATION_STD,
            normalization_scale=1.0 / self.cfg.TRANSFORMATION.NORMALIZATION_SCALE,
            h_flip=0,
            t_crop=False,
            rea=False,
            **self.cfg.EXECUTION.DATAREADER.GENERATOR_ARGS
        )
        self.test_generator.setup(
            crawler_instance,
            mode="test",
            batch_size=self.cfg.TRANSFORMATION.BATCH_SIZE,
            workers=self.cfg.TRANSFORMATION.WORKERS,
            **self.cfg.EXECUTION.DATAREADER.DATASET_ARGS
        )
        self.logger.info("Generated validation data/query generator")

    def buildSaveMetadata(self) -> SaveMetadata:
        """Builds a `SaveMetadata` object containing the model save paths, logger paths, and any other information about saving.
        """
        return SaveMetadata(self.cfg)

    def makeSaveDirectories(self):
        """Creates the save directories for logs, model configuration, metadata, and model save files
        """
        os.makedirs(self.saveMetadata.MODEL_SAVE_FOLDER, exist_ok=True)

    def buildLogger(self, logger: logging.Logger = None) -> logging.Logger:
        """Builds a new logger or adds the correct file and stream handlers to 
        existing logger if it does not already have them. 

        Args:
            logger (logging.Logger, optional): A logger. Defaults to None.

        Returns:
            logging.Logger: A logger with file and stream handlers.
        """
        loggerGiven = True
        if logger is None:
            logger = logging.Logger(self.saveMetadata.MODEL_SAVE_FOLDER)
            loggerGiven = False

        logger_save_path = os.path.join(
            self.saveMetadata.MODEL_SAVE_FOLDER, self.saveMetadata.LOGGER_SAVE_NAME
        )
        # Check for backup logger
        if self.drive_backup:
            backup_logger = os.path.join(
                self.saveMetadata.CHECKPOINT_DIRECTORY,
                self.saveMetadata.LOGGER_SAVE_NAME,
            )
            if os.path.exists(backup_logger):
                print(
                    "Existing log file exists at network backup %s. Will attempt to copy to local directory %s."
                    % (backup_logger, self.saveMetadata.MODEL_SAVE_FOLDER)
                )
                shutil.copy2(backup_logger, self.saveMetadata.MODEL_SAVE_FOLDER)
        if os.path.exists(logger_save_path):
            print(
                "Log file exists at %s. Will attempt to append there."
                % logger_save_path
            )

        streamhandler = False
        filehandler = False

        if logger.hasHandlers():
            for handler in logger.handlers():
                if isinstance(handler, logging.StreamHandler):
                    streamhandler = True
                if isinstance(handler, logging.FileHandler):
                    if handler.baseFilename == os.path.abspath(logger_save_path):
                        filehandler = True

        if not loggerGiven:
            logger.setLevel(logging.DEBUG)

        if not filehandler:
            fh = logging.FileHandler(logger_save_path)
            fh.setLevel(self.logLevels[self.verbose])
            formatter = logging.Formatter("%(asctime)s %(message)s", datefmt="%H:%M:%S")
            fh.setFormatter(formatter)
            logger.addHandler(fh)

        if not streamhandler:
            cs = logging.StreamHandler()
            cs.setLevel(self.logLevels[self.verbose])
            cs.setFormatter(
                logging.Formatter("%(asctime)s %(message)s", datefmt="%H:%M:%S")
            )
            logger.addHandler(cs)

        return logger

    def log(self, message: str, verbose: int = 3):
        """Logs a message. TODO needs to be fixed.

        Args:
            message (str): Message to log
            verbose (int, optional): Logging verbosity. Defaults to 3.
        """
        self.logger.log(self.logLevels[verbose], message)

    def printConfiguration(self):
        """Prints the EdnaML configuration
        """
        self.logger.info("*" * 40)
        self.logger.info("")
        self.logger.info("")
        self.logger.info("Using the following configuration:")
        self.logger.info(self.cfg.export("yaml", indent=4))
        self.logger.info("")
        self.logger.info("")
        self.logger.info("*" * 40)



import os
import random, math


class Cars196DataCrawler:
    def __init__(
        self,
        data_folder="Cars196",
        train_folder="cars_train",
        test_folder="cars_test",
        query_folder="",
        **kwargs
    ):
        self.metadata = {}

        self.data_folder = data_folder
        self.train_folder = os.path.join(self.data_folder, train_folder)
        self.test_folder = os.path.join(self.data_folder, test_folder)
        self.query_folder = os.path.join(self.data_folder, query_folder)

        self.logger = kwargs.get("logger")

        self.__verify(self.data_folder)
        self.__verify(self.train_folder)
        self.__verify(self.test_folder)
        self.__verify(self.query_folder)

        self.crawl()

    def __verify(self, folder):
        if not os.path.exists(folder):
            raise IOError(
                "Folder {data_folder} does not exist".format(data_folder=folder)
            )
        else:
            self.logger.info("Found {data_folder}".format(data_folder=folder))

    def crawl(self,):
        self.metadata["train"], self.metadata["test"], self.metadata["query"] = (
            {},
            {},
            {},
        )
        self.__crawl(
            os.path.join(self.data_folder, "cars_train_annos.mat"), self.train_folder
        )

        self.logger.info(
            "Train\tPIDS: {:6d}\tCIDS: {:6d}\tIMGS: {:8d}".format(
                self.metadata["train"]["pids"],
                self.metadata["train"]["cids"],
                self.metadata["train"]["imgs"],
            )
        )
        self.logger.info(
            "Test \tPIDS: {:6d}\tCIDS: {:6d}\tIMGS: {:8d}".format(
                self.metadata["test"]["pids"],
                self.metadata["test"]["cids"],
                self.metadata["test"]["imgs"],
            )
        )
        self.logger.info(
            "Query\tPIDS: {:6d}\tCIDS: {:6d}\tIMGS: {:8d}".format(
                self.metadata["query"]["pids"],
                self.metadata["query"]["cids"],
                self.metadata["query"]["imgs"],
            )
        )

    def __crawl(self, crawler_annotations_file, folder):
        from scipy.io import loadmat

        matfile = loadmat(crawler_annotations_file)
        # matlab file is car_train_annos.mat or car_test_annos.mat
        crawler = [
            (os.path.join(folder, str(item[5][0])), item[4][0][0] - 1, 0)
            for item in matfile["annotations"][0]
        ]
        # crawler is a list of 3-tuples. Length of N=number of images.
        # Each tuple is (path/to/image, PID, CID)   PID --> class (from 0-195), CID --> 0
        # CID is a holdover from other crawlers used for re-id task, where cid, or camera-id is required. To maintain compatibility with SequencedGenerator (which expects CID) until I write a generator for Cars196
        # PID is similarity from person-reid, where PID stands for person ID. In this case, it is a unique class

        train_crawler = [item for item in crawler if item[1] in range(0, 98)]
        test_crawler = [item for item in crawler if item[1] in range(98, 196)]

        random.shuffle(train_crawler)
        split = 0.7
        split_idx = math.ceil(split * len(train_crawler))

        self.metadata["test"]["crawl"] = train_crawler[split_idx:]
        self.metadata["test"]["pids"] = 98
        self.metadata["test"]["cids"] = 1
        self.metadata["test"]["imgs"] = len(self.metadata["test"]["crawl"])
        train_crawler = train_crawler[:split_idx]

        self.metadata["query"]["crawl"] = test_crawler
        self.metadata["query"]["pids"] = 98
        self.metadata["query"]["cids"] = 1
        self.metadata["query"]["imgs"] = len(self.metadata["query"]["crawl"])

        self.metadata["train"]["crawl"] = train_crawler[:split_idx]
        self.metadata["train"]["pids"] = 98
        self.metadata["train"]["cids"] = 1
        self.metadata["train"]["imgs"] = len(train_crawler)



import sys, requests

# from https://sumit-ghosh.com/articles/python-download-progress-bar/
def download(file_, url):
    sys.stdout.write("Downloading %s from %s" % (file_, url))
    with open(file_, "wb") as f:
        response = requests.get(url, stream=True)
        all_data = response.headers.get("content-length")
        if all_data is None:
            f.write(response.content)
        else:
            all_data = int(all_data)
            progress = 0
            for chunk in response.iter_content(
                chunk_size=max(int(all_data / 1000), 1024 * 1024)
            ):
                progress += len(chunk)
                f.write(chunk)
                sys.stdout.write(
                    "\r{}/{} bytes [{}{}]".format(
                        all_data,
                        progress,
                        "█" * int(100 * progress / all_data),
                        "." * (100 - int(100 * progress / all_data)),
                    )
                )
                sys.stdout.flush()
    sys.stdout.write("\nDownload of %s to %s completed\n" % (file_, url))



from torch import nn

from ednaml.utils.blocks import ChannelAttention
from ednaml.utils.blocks import DenseAttention
from ednaml.utils.blocks import InputAttention
from ednaml.utils.blocks import SpatialAttention


class ResnetBottleneck(nn.Module):
    expansion = 4

    def __init__(
        self,
        inplanes,
        planes,
        stride=1,
        downsample=None,
        groups=1,
        base_width=64,
        dilation=1,
        norm_layer=None,
        attention=None,
        input_attention=False,
        part_attention=False,
    ):
        super(ResnetBottleneck, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        width = int(planes * (base_width / 64.0)) * groups

        self.conv1 = nn.Conv2d(inplanes, width, kernel_size=1, bias=False, stride=1)
        self.bn1 = norm_layer(width)
        self.conv2 = nn.Conv2d(
            width,
            width,
            kernel_size=3,
            stride=stride,
            padding=dilation,
            bias=False,
            groups=groups,
            dilation=dilation,
        )
        self.bn2 = norm_layer(width)
        self.conv3 = nn.Conv2d(
            width, planes * self.expansion, kernel_size=1, bias=False, stride=1
        )
        self.bn3 = norm_layer(planes * self.expansion)
        self.relu = nn.ReLU(inplace=True)

        if input_attention:
            self.input_attention = InputAttention(planes)
        else:
            self.input_attention = None

        if attention is None:
            self.ca = None
            self.sa = None
        elif attention == "cbam":
            self.sa = SpatialAttention(kernel_size=3)
            self.ca = ChannelAttention(planes * self.expansion)
        elif attention == "dbam":
            self.ca = ChannelAttention(planes)
            self.sa = DenseAttention(planes)
        else:
            raise NotImplementedError()

        if part_attention:
            self.p_sa = DenseAttention(planes=planes * self.expansion)
            self.p_ca = ChannelAttention(planes * self.expansion)
        else:
            self.p_ca = None
            self.p_sa = None

        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        if self.input_attention is not None:
            x = self.input_attention(x) * x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.ca is not None:
            out = self.ca(out) * out
            out = self.sa(out) * out

        if self.downsample is not None:
            identity = self.downsample(x)

        p_out = out
        part_mask = None
        if self.p_ca is not None:  # Get part attention
            p_out = self.p_sa(p_out) * p_out
            #            p_out = self.p_ca(p_out) * p_out
            p_out = self.relu(p_out)
            part_mask = self.p_ca(p_out)

        out = out + identity
        out = self.relu(out)

        if self.p_ca is not None:  # Concat part attention
            # out = torch.cat([p_out[:,p_out.shape[1]//2:,:,:],out[:,:p_out.shape[1]//2,:,:]],dim=1)
            out = (part_mask * p_out) + ((1 - part_mask) * out)
        return out



from torch import nn

from ednaml.utils.blocks import ChannelAttention
from ednaml.utils.blocks import DenseAttention
from ednaml.utils.blocks import InputAttention
from ednaml.utils.blocks import SpatialAttention


class ResnetBasicBlock(nn.Module):
    """ResNet component block for R18, R34. Useful when there are not too many layers
    to deal with complicated matrix multiplications.

    Raises:
        ValueError: When groups!=1 or base_width!=64
        NotImplementedError: When dilation>1 or attention not cbam or dbam
    """

    expansion = 1

    def __init__(
        self,
        inplanes: int,
        planes: int,
        stride: int = 1,
        downsample=None,
        groups: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        norm_layer=nn.BatchNorm2d,
        attention: str = None,
        input_attention: bool = False,
        part_attention: bool = False,
    ):
        """Sets up the ResNet BasicBlock

        Args:
            inplanes (int): Depth of input
            planes (int): Depth of output
            stride (int, optional): Convolutional stride parameter. Defaults to 1.
            downsample (Union[int,None], optional): Whether to downsample images for skip connection. Defaults to None.
            groups (int, optional): <>. Defaults to 1.
            base_width (int, optional): <>. Defaults to 64.
            dilation (int, optional): Convolutional dilation parameters. Defaults to 1.
            norm_layer (Union[nn.GroupNorm,nn.modules.batchnorm._NormBase,None], optional): Normalization layer throughout Block. Defaults to nn.BatchNorm2d.
            attention (str, optional): Attention type: CBAM or DBAM. Defaults to None.
            input_attention (bool, optional): Whether to include the input attention module. Defaults to False.
            part_attention (bool, optional): Whether to include the part (local) attention module. Defaults to False.
        """
        super(ResnetBasicBlock, self).__init__()
        # Verify some base parameters
        if groups != 1 or base_width != 64:
            raise ValueError("BasicBlock only supports groups=1 and base_width=64")
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        if attention is not None and attention not in ["cbam", "dbam"]:
            raise ValueError(
                "attention parameter is unsupported value %s. Use one of 'cbam','dbam'."
                % attention
            )
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        # This builds the core layers in the BasicBlock
        self.conv1 = nn.Conv2d(
            inplanes,
            planes,
            kernel_size=3,
            stride=stride,
            padding=1,
            bias=False,
            groups=1,
            dilation=1,
        )
        self.bn1 = norm_layer(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = nn.Conv2d(
            planes,
            planes,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=False,
            groups=1,
            dilation=1,
        )
        self.bn2 = norm_layer(planes)

        # Sets up input attention
        self.input_attention = None
        if input_attention:
            self.input_attention = InputAttention(planes)

        # Sets up CBAM or DBAM attention
        self.ca = None
        self.sa = None
        if attention == "cbam":
            self.ca = ChannelAttention(planes)
            self.sa = SpatialAttention(kernel_size=3)
        if attention == "dbam":
            self.ca = ChannelAttention(planes)
            self.sa = DenseAttention(planes)

        # Sets up local attention
        self.p_ca = None
        self.p_sa = None
        if part_attention:
            self.p_sa = DenseAttention(planes=planes * self.expansion)
            self.p_ca = ChannelAttention(planes * self.expansion)

        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        if self.input_attention is not None:
            x = self.input_attention(x) * x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.ca is not None:
            out = self.ca(out) * out
            out = self.sa(out) * out

        if self.downsample is not None:
            identity = self.downsample(x)

        p_out = out
        part_mask = None
        if self.p_ca is not None:  # Get part attention
            p_out = self.p_sa(p_out) * p_out
            #            p_out = self.p_ca(p_out) * p_out
            p_out = self.relu(p_out)
            part_mask = self.p_ca(p_out)

        out = out + identity
        out = self.relu(out)

        if self.p_ca is not None:  # Concat part attention
            # out = torch.cat([p_out[:,p_out.shape[1]//2:,:,:],out[:,:p_out.shape[1]//2,:,:]],dim=1)
            out = (part_mask * p_out) + ((1 - part_mask) * out)
        return out



import torch
import torch.nn.functional as F
from ednaml.loss import Loss


class ClusterLoss(Loss):
    def __init__(self, margin: float, instances: int, images_per_instance: int):
        """ ClusterLoss from Cluster Loss for Person Re-Identification

        Args:
            margin (float): margin for cluster loss
            instances (int): Number of unique ids in a batch
            images_per_instance (int): Number of images per instance in batch. Should be batch_size / instances
        """
        super(ClusterLoss, self).__init__()
        self.margin = margin
        self.ids_per_batch = instances
        self.imgs_per_id = images_per_instance

    def _euclidean_dist(self, x, y):
        """
        Args:
          x: pytorch Variable, with shape [m, d]
          y: pytorch Variable, with shape [n, d]
        Returns:
          dist: pytorch Variable, with shape [m, n]
        """
        m, n = x.size(0), y.size(0)
        xx = torch.pow(x, 2).sum(1, keepdim=True).expand(m, n)
        yy = torch.pow(y, 2).sum(1, keepdim=True).expand(n, m).t()
        dist = xx + yy
        dist.addmm_(1, -2, x, y.t())
        dist = dist.clamp(min=1e-12).sqrt()  # for numerical stability
        return dist

    def _cluster_loss(self, features, targets, ids_per_batch=32, imgs_per_id=4):
        """
        Args:
            features: prediction matrix (before softmax) with shape (batch_size, feature_dim)
            targets: ground truth labels with shape (batch_size)
            ids_per_batch: num of different ids per batch
            imgs_per_id: num of images per id
        Return:
             cluster_loss
        """

        unique_labels = targets.cpu().unique().cuda()

        inter_min_distance = torch.zeros(unique_labels.size(0))
        intra_max_distance = torch.zeros(unique_labels.size(0))
        center_features = torch.zeros(unique_labels.size(0), features.size(1))

        inter_min_distance = inter_min_distance.cuda()
        intra_max_distance = intra_max_distance.cuda()
        center_features = center_features.cuda()

        index = torch.range(0, unique_labels.size(0) - 1)
        for i in range(unique_labels.size(0)):
            label = unique_labels[i]
            same_class_features = features[targets == label]
            center_features[i] = same_class_features.mean(dim=0)
            intra_class_distance = self._euclidean_dist(
                center_features[index == i], same_class_features
            )
            # print('intra_class_distance', intra_class_distance)
            intra_max_distance[i] = intra_class_distance.max()
        # print('intra_max_distance:', intra_max_distance)

        for i in range(unique_labels.size(0)):
            inter_class_distance = self._euclidean_dist(
                center_features[index == i], center_features[index != i]
            )
            # print('inter_class_distance', inter_class_distance)
            inter_min_distance[i] = inter_class_distance.min()
        #  print('inter_min_distance:', inter_min_distance)
        cluster_loss = torch.mean(
            torch.relu(intra_max_distance - inter_min_distance + self.margin)
        )
        return cluster_loss, intra_max_distance, inter_min_distance

    def forward(self, features, labels):
        """
        Args:
            features: prediction matrix (before softmax) with shape (batch_size, feature_dim)
            labels: ground truth labels with shape (batch_size)

        Return:
             cluster_loss
        """
        cluster_loss, cluster_dist_ap, cluster_dist_an = self._cluster_loss(
            features, labels, self.ids_per_batch, self.imgs_per_id
        )
        return cluster_loss



from ednaml.datareaders import DataReader
from ednaml.crawlers import CoLabelCompCarsCrawler
from ednaml.generators.ClassificationGenerator import (
    ClassificationDataset,
    ClassificationGenerator,
)


class CompCars(DataReader):
    name: str = "CompCars"
    CRAWLER = CoLabelCompCarsCrawler
    DATASET = ClassificationDataset
    GENERATOR = ClassificationGenerator

    def __init__(self):
        super().__init__()



import importlib
import torch
from ednaml.loss.builders import LossBuilder
from ednaml.optimizer import BaseOptimizer


class StandardLossOptimizer(BaseOptimizer):
    """ Optimizer for Loss Functions

    This sets up optimizer for the loss functions in a LossBuilder. Not applicable everywhere. Not used everywhere because most losses will have zero differentiable parametersif 
    
    However, it will be useful for losses like the ProxyNCA, which need to learn proxies during training.

    """

    def __init__(
        self,
        name,
        optimizer,
        base_lr,
        lr_bias,
        gpus,
        weight_decay,
        weight_bias,
        opt_kwargs,
    ):
        """ Initializes the optimizer builder.

        Args:
        base_lr (float): Base learning rate for optimizer
        lr_bias (float): Multiplicative factor for bias parameters
        gpus (int): Number of GPUs for lr scaling
        weight_decay (float): Weight decay for decoupled weight decay optimizers like AdamW
        weight_bias (float): Multiplicative factor for bias parameters in weight decay optimizers

        Methods:
        build:  builds an optimizer given optimizer name and torch model

        """
        super(StandardLossOptimizer, self).__init__(
            name,
            optimizer,
            base_lr,
            lr_bias,
            gpus,
            weight_decay,
            weight_bias,
            opt_kwargs,
        )

    def build(self, loss_builder: LossBuilder) -> torch.optim.Optimizer:
        """ Builds an optimizer.

        Args:
        loss_builder (loss.builders.LossBuilder): A LossBuilder object
        name (str): name of torch.optim object to build
        kwargs (dict): any parameters that need to be passed into the optimizer

        Returns:
        torch.optim object

        """
        params = []
        for key, value in loss_builder.named_parameters():
            if value.requires_grad:
                # if "bias" in key:
                #    learning_rate = self.base_lr * self.lr_bias
                #    weight_decay = self.weight_decay * self.weight_bias
                # else:
                learning_rate = self.base_lr * self.gpus
                weight_decay = self.weight_decay
                params += [
                    {
                        "params": [value],
                        "lr": learning_rate,
                        "weight_decay": weight_decay,
                    }
                ]
        if len(params) == 0:
            return None
        optimizer = importlib.import_module(self.optimizer, package="torch.optim")
        optimizer = optimizer(params, **self.kwargs)
        return optimizer



from ednaml.utils.LabelMetadata import LabelMetadata
from ednaml.models.ModelAbstract import ModelAbstract


def classification_model_builder(
    arch,
    base,
    weights=None,
    normalization=None,
    metadata: LabelMetadata = None,
    **kwargs
):
    """Corroborative/Colaborative/Complementary Labeler Model Builder

    This builds a model for colabeler. Refer to paper [] for general construction. The model contains:
        * Architecture core
        * Convolutional attention
        * Spatial average pooling
        * FC-Layer for embedding dimensionality change
        * Normalization layer
        * Softmax FC Layers

    Args:
        arch (str): Ther architecture to use. The string "Base" is added after architecture (e.g. "Resnet", "Inception"). Only "Resnet" is currently supported.
        base (str): The architecture subtype, e.g. "resnet50", "resnet18"
        weights (str): Local path to weights file for the architecture core, e.g. pretrained resnet50 weights path.
        normalization (str): Normalization layer for reid-model. Can be None. Supported normalizations: ["bn", "l2", "in", "gn", "ln"]
    
    KWARGS (dict):
        embedding_dimension (int): Dimensions for the feature embedding. Leave empty if feature dimensions should be same as architecture core output (e.g. resnet50 base model has 2048-dim feature outputs). If providing a value, it should be less than the architecture core's base feature dimensions
        metadata: dict of class-name: num-classes to infer size of softmax dimensions.


    Returns:
        Torch Model: A Torch classification model

    """
    if arch != "ClassificationResnet":
        raise NotImplementedError()
    archbase = __import__("models." + arch, fromlist=[arch])
    archbase = getattr(archbase, arch)

    # Extract the dimensions from the classes metadata provided by EdnaML
    kwargs["softmax_dimensions"] = metadata.getLabelDimensions()

    model = archbase(
        base=base,
        weights=weights,
        normalization=normalization,
        metadata=metadata,
        **kwargs
    )
    return model


def multiclassification_model_builder(
    arch, base, weights=None, normalization=None, metadata=None, **kwargs
):
    """Multiclassification model builder. This builds a model with a single backbone, and multiple classification FC layers.

    The model contains:
        * Architecture core
        * Convolutional attention
        * Spatial average pooling
        * Normalization layer
        * Softmax FC Layers

    Args:
        arch (str): Ther architecture to use. The string "Base" is added after architecture (e.g. "MultiClassificationResnet", "MultiClassificationInception"). Only "MultiClassificationResnet" is currently supported.
        base (str): The architecture subtype, e.g. "resnet50", "resnet18"
        weights (str): Local path to weights file for the architecture core, e.g. pretrained resnet50 weights path.
        normalization (str): Normalization layer for reid-model. Can be None. Supported normalizations: ["bn", "l2", "in", "gn", "ln"]
        metadata (Dict[str, int]): Label names to num-classes dictionary

    Kwargs:
        - `number_outputs`: This is the number of classification outputs for the model
        - `outputs`: A list, each element is the i-th output's metadata
            - `dimensions`: This is the number of classes for this output. This can be left blank if you want EdnaML to infer the size from the `LABEL` parameter below. 
            - `name`: This is the name of this output. This is used by model_builder to keep track of output names and labels. If left blank, EdnaML will automatically name all outputs. Outputs are named because multiple outputs can track the same class, sometimes, e.g. in a contrastive non-weight sharing setting
            - `label`: This is the name of the label this output is tracking. <span style="color:magenta; font-weight:bold">THIS SHOULD CORRESPOND EXACTLY WITH `DATASET_ARGS.classificationclass`</span> labels. 
    Returns:
        Torch Model: A Torch multiclassification model

    """

    # make the MultiClassificationResnet, with resnet base, with multiple output fc layers
    if arch != "MultiClassificationResnet":
        raise NotImplementedError()
    archbase = __import__("models." + arch, fromlist=[arch])
    archbase = getattr(archbase, arch)

    model = archbase(
        base=base,
        weights=weights,
        normalization=normalization,
        metadata=metadata,
        **kwargs
    )
    return model


def multibranch_model_builder(
    arch, base, weights=None, normalization=None, metadata=None, **kwargs
):
    """Multibranch model builder builds a model with multiple branches, each with their set of outputs.

    The model contains:
        * Architecture core
        * Convolutional attention
        * Spatial average pooling
        * Normalization layer
        * Softmax FC Layers

    Args:
        arch (str): Ther architecture to use. The string "Base" is added after architecture (e.g. "MultiClassificationResnet", "MultiClassificationInception"). Only "MultiClassificationResnet" is currently supported.
        base (str): The architecture subtype, e.g. "resnet50", "resnet18"
        weights (str): Local path to weights file for the architecture core, e.g. pretrained resnet50 weights path.
        normalization (str): Normalization layer for reid-model. Can be None. Supported normalizations: ["bn", "l2", "in", "gn", "ln"]
        metadata (Dict[str, int]): Label names to num-classes dictionary

    Kwargs:
        - `number_branches`: This is the number of branches for the model
        - `branches`: A list, each element is the i-th branch's metadata
            - `name`: This is the name of this branch. This is used by model_builder to keep track of output names and labels. 
            - `number_outputs`: This is the number of classification outputs for this branch
            - `outputs`: A list, each element is the j-th output's metadata for this branch
                - `dimensions`: This is the number of classes for this output. This can be left blank if you want EdnaML to infer the size from the `LABEL` parameter below. 
                - `name`: This is the name of this output. 
                - `label`: This is the name of the label this output is tracking. <span style="color:magenta; font-weight:bold">THIS SHOULD CORRESPOND EXACTLY WITH `DATASET_ARGS.classificationclass`</span> labels.
        - `fuse`: **Bool**. Whether branch outputs are going to be fused
        - `fuse_outputs`: List of output names (not branch names) that are fused
        - `fuse_dimensions`: The dimensions of the fused output. If left blank, EdnaML will infer from `fuse_label`
        - `fuse_label`: The label that tracks the fused output 

    Returns:
        Torch Model: A Torch Re-ID model

    """

    # make the MultiClassificationResnet, with resnet base, with multiple output fc layers
    if arch != "MultiBranchResnet":
        raise NotImplementedError()
    archbase = __import__("models." + arch, fromlist=[arch])
    archbase = getattr(archbase, arch)

    model = archbase(
        base=base,
        weights=weights,
        normalization=normalization,
        metadata=metadata,
        **kwargs
    )
    return model



import torch
from torch import nn
from ednaml.utils.blocks import ChannelAttention, DenseAttention, InputAttention


class SELayer(nn.Module):
    def __init__(self, inplanes, isTensor=True):
        super(SELayer, self).__init__()
        if isTensor:
            # if the input is (N, C, H, W)
            self.SE_opr = nn.Sequential(
                nn.AdaptiveAvgPool2d(1),
                nn.Conv2d(inplanes, inplanes // 4, kernel_size=1, stride=1, bias=False),
                nn.BatchNorm2d(inplanes // 4),
                nn.ReLU(inplace=True),
                nn.Conv2d(inplanes // 4, inplanes, kernel_size=1, stride=1, bias=False),
            )
        else:
            # if the input is (N, C)
            self.SE_opr = nn.Sequential(
                nn.AdaptiveAvgPool2d(1),
                nn.Linear(inplanes, inplanes // 4, bias=False),
                nn.BatchNorm1d(inplanes // 4),
                nn.ReLU(inplace=True),
                nn.Linear(inplanes // 4, inplanes, bias=False),
            )

    def forward(self, x):
        atten = self.SE_opr(x)
        atten = torch.clamp(atten + 3, 0, 6) / 6
        return x * atten


class HS(nn.Module):
    def __init__(self):
        super(HS, self).__init__()

    def forward(self, inputs):
        clip = torch.clamp(inputs + 3, 0, 6) / 6
        return inputs * clip


class Shufflenet(nn.Module):
    def __init__(
        self, inp, oup, base_mid_channels, *, ksize, stride, activation, useSE
    ):
        super(Shufflenet, self).__init__()
        self.stride = stride
        assert stride in [1, 2]
        assert ksize in [3, 5, 7]
        assert base_mid_channels == oup // 2

        self.base_mid_channel = base_mid_channels
        self.ksize = ksize
        pad = ksize // 2
        self.pad = pad
        self.inp = inp

        outputs = oup - inp

        branch_main = nn.ModuleList(
            [
                # pw
                nn.Conv2d(inp, base_mid_channels, 1, 1, 0, bias=False),
                nn.BatchNorm2d(base_mid_channels),
                None,
                # dw
                nn.Conv2d(
                    base_mid_channels,
                    base_mid_channels,
                    ksize,
                    stride,
                    pad,
                    groups=base_mid_channels,
                    bias=False,
                ),
                nn.BatchNorm2d(base_mid_channels),
                # pw-linear
                nn.Conv2d(base_mid_channels, outputs, 1, 1, 0, bias=False),
                nn.BatchNorm2d(outputs),
                None,
            ]
        )
        if activation == "ReLU":
            assert useSE == False
            """This model should not have SE with ReLU"""
            branch_main[2] = nn.ReLU(inplace=True)
            branch_main[-1] = nn.ReLU(inplace=True)
        else:
            branch_main[2] = HS()
            branch_main[-1] = HS()
            if useSE:
                branch_main.append(SELayer(outputs))
        self.branch_main = nn.Sequential(*branch_main)

        if stride == 2:
            branch_proj = nn.ModuleList(
                [
                    # dw
                    nn.Conv2d(inp, inp, ksize, stride, pad, groups=inp, bias=False),
                    nn.BatchNorm2d(inp),
                    # pw-linear
                    nn.Conv2d(inp, inp, 1, 1, 0, bias=False),
                    nn.BatchNorm2d(inp),
                    None,
                ]
            )
            if activation == "ReLU":
                branch_proj[-1] = nn.ReLU(inplace=True)
            else:
                branch_proj[-1] = HS()
            self.branch_proj = nn.Sequential(*branch_proj)
        else:
            self.branch_proj = None

    def forward(self, old_x):
        if self.stride == 1:
            x_proj, x = channel_shuffle(old_x)
            return torch.cat((x_proj, self.branch_main(x)), 1)
        elif self.stride == 2:
            x_proj = old_x
            x = old_x
            return torch.cat((self.branch_proj(x_proj), self.branch_main(x)), 1)


class Shuffle_Xception(nn.Module):
    def __init__(self, inp, oup, base_mid_channels, *, stride, activation, useSE):
        super(Shuffle_Xception, self).__init__()

        assert stride in [1, 2]
        assert base_mid_channels == oup // 2

        self.base_mid_channel = base_mid_channels
        self.stride = stride
        self.ksize = 3
        self.pad = 1
        self.inp = inp
        outputs = oup - inp

        branch_main = nn.ModuleList(
            [
                # dw
                nn.Conv2d(inp, inp, 3, stride, 1, groups=inp, bias=False),
                nn.BatchNorm2d(inp),
                # pw
                nn.Conv2d(inp, base_mid_channels, 1, 1, 0, bias=False),
                nn.BatchNorm2d(base_mid_channels),
                None,
                # dw
                nn.Conv2d(
                    base_mid_channels,
                    base_mid_channels,
                    3,
                    stride,
                    1,
                    groups=base_mid_channels,
                    bias=False,
                ),
                nn.BatchNorm2d(base_mid_channels),
                # pw
                nn.Conv2d(base_mid_channels, base_mid_channels, 1, 1, 0, bias=False),
                nn.BatchNorm2d(base_mid_channels),
                None,
                # dw
                nn.Conv2d(
                    base_mid_channels,
                    base_mid_channels,
                    3,
                    stride,
                    1,
                    groups=base_mid_channels,
                    bias=False,
                ),
                nn.BatchNorm2d(base_mid_channels),
                # pw
                nn.Conv2d(base_mid_channels, outputs, 1, 1, 0, bias=False),
                nn.BatchNorm2d(outputs),
                None,
            ]
        )

        if activation == "ReLU":
            branch_main[4] = nn.ReLU(inplace=True)
            branch_main[9] = nn.ReLU(inplace=True)
            branch_main[14] = nn.ReLU(inplace=True)
        else:
            branch_main[4] = HS()
            branch_main[9] = HS()
            branch_main[14] = HS()
        assert None not in branch_main

        if useSE:
            assert activation != "ReLU"
            branch_main.append(SELayer(outputs))

        self.branch_main = nn.Sequential(*branch_main)

        if self.stride == 2:
            branch_proj = nn.ModuleList(
                [
                    # dw
                    nn.Conv2d(inp, inp, 3, stride, 1, groups=inp, bias=False),
                    nn.BatchNorm2d(inp),
                    # pw-linear
                    nn.Conv2d(inp, inp, 1, 1, 0, bias=False),
                    nn.BatchNorm2d(inp),
                    None,
                ]
            )
            if activation == "ReLU":
                branch_proj[-1] = nn.ReLU(inplace=True)
            else:
                branch_proj[-1] = HS()
            self.branch_proj = nn.Sequential(*branch_proj)

    def forward(self, old_x):
        if self.stride == 1:
            x_proj, x = channel_shuffle(old_x)
            return torch.cat((x_proj, self.branch_main(x)), 1)
        elif self.stride == 2:
            x_proj = old_x
            x = old_x
            return torch.cat((self.branch_proj(x_proj), self.branch_main(x)), 1)


def channel_shuffle(x):
    batchsize, num_channels, height, width = x.data.size()
    assert num_channels % 4 == 0
    x = x.reshape(batchsize * num_channels // 2, 2, height * width)
    x = x.permute(1, 0, 2)
    x = x.reshape(2, -1, num_channels // 2, height, width)
    return x[0], x[1]


class ShuffleNetV2_Plus(nn.Module):
    def __init__(
        self,
        input_size=224,
        architecture=None,
        model_size="Large",
        ia_attention=True,
        part_attention=True,
    ):
        super(ShuffleNetV2_Plus, self).__init__()

        assert input_size % 32 == 0
        assert architecture is not None

        self.stage_repeats = [4, 4, 8, 4]
        if model_size == "Large":
            self.stage_out_channels = [-1, 16, 68, 168, 336, 672, 1280]
        elif model_size == "Medium":
            self.stage_out_channels = [-1, 16, 48, 128, 256, 512, 1280]
        elif model_size == "Small":
            self.stage_out_channels = [-1, 16, 36, 104, 208, 416, 1280]
        else:
            raise NotImplementedError

        # building first layer
        input_channel = self.stage_out_channels[1]
        self.first_conv = nn.Sequential(
            nn.Conv2d(3, input_channel, 3, 2, 1, bias=False),
            nn.BatchNorm2d(input_channel),
            HS(),
        )

        self.features = nn.ModuleList([])
        archIndex = 0
        for idxstage in range(len(self.stage_repeats)):
            numrepeat = self.stage_repeats[idxstage]
            output_channel = self.stage_out_channels[idxstage + 2]

            activation = "HS" if idxstage >= 1 else "ReLU"
            useSE = "True" if idxstage >= 2 else False

            for i in range(numrepeat):
                if i == 0:
                    inp, outp, stride = input_channel, output_channel, 2
                else:
                    inp, outp, stride = input_channel // 2, output_channel, 1

                blockIndex = architecture[archIndex]
                archIndex += 1
                if blockIndex == 0:
                    # print('Shuffle3x3')
                    self.features.append(
                        Shufflenet(
                            inp,
                            outp,
                            base_mid_channels=outp // 2,
                            ksize=3,
                            stride=stride,
                            activation=activation,
                            useSE=useSE,
                        )
                    )
                elif blockIndex == 1:
                    # print('Shuffle5x5')
                    self.features.append(
                        Shufflenet(
                            inp,
                            outp,
                            base_mid_channels=outp // 2,
                            ksize=5,
                            stride=stride,
                            activation=activation,
                            useSE=useSE,
                        )
                    )
                elif blockIndex == 2:
                    # print('Shuffle7x7')
                    self.features.append(
                        Shufflenet(
                            inp,
                            outp,
                            base_mid_channels=outp // 2,
                            ksize=7,
                            stride=stride,
                            activation=activation,
                            useSE=useSE,
                        )
                    )
                elif blockIndex == 3:
                    # print('Xception')
                    self.features.append(
                        Shuffle_Xception(
                            inp,
                            outp,
                            base_mid_channels=outp // 2,
                            stride=stride,
                            activation=activation,
                            useSE=useSE,
                        )
                    )
                else:
                    raise NotImplementedError
                input_channel = output_channel
        assert archIndex == len(architecture)
        # self.features = nn.Sequential(*self.features)      # manually do it for attention....

        self.conv_last = nn.Sequential(
            nn.Conv2d(input_channel, 1280, 1, 1, 0, bias=False),
            nn.BatchNorm2d(1280),
            HS(),
        )
        self.globalpool = nn.AvgPool2d(7)

        """ Don't need these.
        self.LastSE = SELayer(1280)
        self.fc = nn.Sequential(
            nn.Linear(1280, 1280, bias=False),
            HS(),
        )
        self.dropout = nn.Dropout(0.2)
        self.classifier = nn.Sequential(nn.Linear(1280, n_class, bias=False))
        """
        # Create true IA
        if ia_attention:
            self.ia_attention = InputAttention(
                16
            )  # MAGIC...use from architecture array
        else:
            self.ia_attention = None

        if part_attention:
            self.p_sa = DenseAttention(36)  # MAGIC...use from architecture array
            self.p_ca = ChannelAttention(36)
            self.p_relu = nn.ReLU(inplace=True)
        else:
            self.p_ca = None
            self.p_sa = None
            self.p_relu = None

        self._initialize_weights()

    def forward(self, x):
        x = self.first_conv(x)
        if self.ia_attention is not None:
            x = self.ia_attention(x) * x

        x = self.features[0](x)
        if self.p_ca is not None:
            p_out = self.p_sa(x) * x
            p_out = self.p_relu(p_out)
            part_mask = self.p_ca(p_out)
            x = (part_mask * p_out) + ((1 - part_mask) * x)

        for idx, layer in enumerate(self.features[1:]):
            x = layer(x)

        x = self.conv_last(x)

        x = self.globalpool(x)

        """x = self.LastSE(x)

        x = x.contiguous().view(-1, 1280)

        x = self.fc(x)
        x = self.dropout(x)
        x = self.classifier(x)"""
        return x

    def _initialize_weights(self):
        for name, m in self.named_modules():
            if isinstance(m, nn.Conv2d):
                if "first" in name or "SE" in name:
                    nn.init.normal_(m.weight, 0, 0.01)
                else:
                    nn.init.normal_(m.weight, 0, 1.0 / m.weight.shape[1])
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0.0001)
                nn.init.constant_(m.running_mean, 0)
            elif isinstance(m, nn.BatchNorm1d):
                nn.init.constant_(m.weight, 1)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0.0001)
                nn.init.constant_(m.running_mean, 0)
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

    def load_param(self, weights_path):
        params = torch.load(weights_path)
        if "state_dict" in params:
            # This is shufflenet style, not ours...
            for i in params["state_dict"]:  # ShuffleNet weights specific issue
                if "fc" in i:
                    continue
                # All 'i' in params have "module." in front of their layer names. [7:] gets rid of it for our models...
                if (
                    i[7:] not in self.state_dict()
                    or params["state_dict"][i].shape != self.state_dict()[i[7:]].shape
                ):
                    continue
                self.state_dict()[i[7:]].copy_(params["state_dict"][i])
        else:  # Our style
            for _key in params:
                if (
                    _key not in self.state_dict().keys()
                    or params[_key].shape != self.state_dict()[_key].shape
                ):
                    continue
                self.state_dict()[_key].copy_(params[_key])


def _shufflenetv2_plus(model_size="Small", **kwargs):
    # Set up shufflenet architecture (from https://github.com/megvii-model/ShuffleNet-Series/blob/master/ShuffleNetV2%2B/train.py)
    architecture = [0, 0, 3, 1, 1, 1, 0, 0, 2, 0, 2, 1, 1, 0, 2, 0, 2, 1, 3, 2]
    model = ShuffleNetV2_Plus(
        architecture=architecture, model_size=model_size, **kwargs
    )
    return model


def shufflenetv2_small(**kwargs):
    """ShuffleNetv2-Small model from https://github.com/megvii-model/ShuffleNet-Series/
    """
    return _shufflenetv2_plus(model_size="Small", **kwargs)



import os


class Crawler:
    """The base crawler class. TODO
    """

    def __init__():
        raise NotImplementedError()

    def __verify(self, folder):
        if not os.path.exists(folder):
            raise IOError(
                "Folder {data_folder} does not exist".format(data_folder=folder)
            )
        else:
            self.logger.info("Found {data_folder}".format(data_folder=folder))


# Vehicle Re-ID Crawlers
from ednaml.crawlers.VeRiDataCrawler import VeRiDataCrawler
from ednaml.crawlers.VRICDataCrawler import VRICDataCrawler
from ednaml.crawlers.VehicleIDDataCrawler import VehicleIDDataCrawler

# Carzam crawlers
from ednaml.crawlers.Cars196DataCrawler import Cars196DataCrawler

# Other Crawlers
from ednaml.crawlers.ClassedCrawler import ClassedCrawler


# CoLabelCrawlers
from ednaml.crawlers.CoLabelVehicleColorCrawler import CoLabelVehicleColorCrawler
from ednaml.crawlers.CoLabelIntegratedDatasetCrawler import (
    CoLabelIntegratedDatasetCrawler,
)

KnowledgeIntegratedDatasetCrawler = CoLabelIntegratedDatasetCrawler

from ednaml.crawlers.CoLabelCompCarsCrawler import CoLabelCompCarsCrawler



import torch


def pairwise_distance(a, squared=False, eps=1e-16):
    """Computes the pairwise distance matrix with numerical stability."""
    operand = a.pow(2).sum(dim=1, keepdim=True).expand(a.size(0), -1)
    operandT = torch.t(a).pow(2).sum(dim=0, keepdim=True).expand(a.size(0), -1)
    pairwise_distances_squared = torch.add(operand, operandT) - 2 * (
        torch.mm(a, torch.t(a))
    )
    pairwise_distances_squared = torch.clamp(pairwise_distances_squared, min=0.0)
    error_mask = torch.le(pairwise_distances_squared, 0.0)

    if squared:
        pairwise_distances = pairwise_distances_squared
    else:
        pairwise_distances = torch.sqrt(
            pairwise_distances_squared + error_mask.float() * eps
        )

    # Undo conditionally adding 1e-16.
    pairwise_distances = torch.mul(pairwise_distances, (error_mask == False).float())

    # Explicitly set diagonals to zero.
    mask_offdiagonals = 1 - torch.eye(
        *pairwise_distances.size(), device=pairwise_distances.device
    )
    pairwise_distances = torch.mul(pairwise_distances, mask_offdiagonals)
    return pairwise_distances



from typing import Dict

from ednaml.config.ConfigDefaults import ConfigDefaults


class ModelConfig:
    BUILDER: str
    MODEL_ARCH: str
    MODEL_BASE: str
    MODEL_NORMALIZATION: str
    MODEL_KWARGS: Dict[str, str]

    def __init__(self, model_dict, defaults: ConfigDefaults):
        self.BUILDER = model_dict.get("BUILDER", defaults.BUILDER)
        self.MODEL_ARCH = model_dict.get("MODEL_ARCH", defaults.MODEL_ARCH)
        self.MODEL_BASE = model_dict.get("MODEL_BASE", defaults.MODEL_BASE)
        self.MODEL_NORMALIZATION = model_dict.get(
            "MODEL_NORMALIZATION", defaults.MODEL_NORMALIZATION
        )
        self.MODEL_KWARGS = model_dict.get("MODEL_KWARGS", defaults.MODEL_KWARGS)



from torch import nn
from ednaml.models.ClassificationResnet import ClassificationResnet


class MultiClassificationResnet(ClassificationResnet):
    """Multiclassification Resnet model, that performs multiple classifications from the same backbone.

    A MulticlassificationResnet model is a base ResNet with multiple FC classification layers.

    Args: (TODO)
        base (str): The architecture base for resnet, i.e. resnet50, resnet18
        weights (str, None): Path to weights file for the architecture base ONLY. If not provided, base initialized with random values.
        normalization (str, None): Can be None, where it is torch's normalization. Else create a normalization layer. Supports: ["bn", "l2", "in", "gn", "ln"]
        metadata (Dict[str:int], None): FC dimensions for each classification, keyed by class names

    Kwargs (MODEL_KWARGS):
        number_outputs (int): Number of different FC layers connected to the feature layer of the backbone.
        softmax_dimensions (List[int]): Classes of each FC layer, in order. Optional. If not provided, `MultiClassificationResnet` will use `output_classnames` and `metadata` to infer dimension size
        output_classnames (List[str]): The name for each output, in order. These should be the same as the label names for the multiple classes.
        labelnames (List[str]): The order of labels provided by the crawler. This is used during model training, where crawler ground truth labels must be matched to model outputs. 
        
        last_stride (int, 1): The final stride parameter for the architecture core. Should be one of 1 or 2.
        attention (str, None): The attention module to use. Only supports ['cbam', 'dbam']
        input_attention (bool, false): Whether to include the IA module
        secondary_attention (List[int], None): Whether to modify CBAM to apply it to specific Resnet basic blocks. None means CBAM is applied to all. Otherwise, CBAM is applied only to the basic blocks provided here in List.
        part_attention (bool): Whether to use local attention

    Default Kwargs (DO NOT CHANGE OR ADD TO MODEL_KWARGS):
        zero_init_residual (bool, false): Whether the final layer uses zero initialization
        top_only (bool, true): Whether to keep only the architecture base without imagenet fully-connected layers (1000 classes)
        num_classes (int, 1000): Number of features in final imagenet FC layer
        groups (int, 1): Used during resnet variants construction
        width_per_group (int, 64): Used during resnet variants construction
        replace_stride_with_dilation (bool, None): Well, replace stride with dilation...
        norm_layer (nn.Module, None): The normalization layer within resnet. Internally defaults to nn.BatchNorm2D

    Methods: 
        forward: Process a batch

    """

    model_name = "MultiClassificationResNet"
    model_arch = "MultiClassificationResNet"
    number_outputs = 1
    output_names = ["out0"]
    softmax_dimensions = [None]
    output_labels = ["color"]
    secondary_outputs = []

    _internal_name_count = 0

    def __init__(
        self, base="resnet50", weights=None, normalization=None, metadata=None, **kwargs
    ):
        """We will inherit the base construction from ClassificationResNet, and modify the softmax head.

        Args:
            base (str, optional): _description_. Defaults to 'resnet50'.
            weights (_type_, optional): _description_. Defaults to None.
            normalization (_type_, optional): _description_. Defaults to None.
            metadata (_type_, optional): _description_. Defaults to None.
        """

        super().__init__(
            base=base,
            weights=weights,
            normalization=normalization,
            metadata=metadata,
            **kwargs
        )

    def model_attributes_setup(self, **kwargs):

        self.embedding_dimensions = kwargs.get("embedding_dimensions", None)
        if self.normalization == "":
            self.normalization = None

        self.number_outputs = kwargs.get("number_outputs", 1)

        outputs = kwargs.get(
            "outputs",
            [
                {
                    "name": self._internal_name_counter(),
                    "label": "color",
                    "dimensions": None,
                }
            ],
        )

        if len(outputs) != self.number_outputs:
            raise ValueError(
                "Mismatch in length of outputs %i and number of outputs %i"
                % (len(outputs), self.number_outputs)
            )

        self.softmax_dimensions = [None] * self.number_outputs
        self.output_names = [None] * self.number_outputs
        self.output_labels = [None] * self.number_outputs

        for idx, output_details in enumerate(outputs):
            self.softmax_dimensions[idx] = output_details.get("dimensions", None)
            self.output_names[idx] = output_details.get(
                "name", self._internal_name_counter()
            )
            self.output_labels[idx] = output_details["label"]
            if self.softmax_dimensions[idx] is None:
                self.softmax_dimensions[idx] = self.metadata.getLabelDimensions(
                    self.output_labels[idx]
                )
        self.base = None
        self.gap = None
        self.emb_linear = None
        self.feat_norm = None
        self.softmax = None

    def _internal_name_counter(self):
        out = "out" + str(self._internal_name_count)
        self._internal_name_count += 1
        return out

    def build_softmax(self, **kwargs):
        """Build the softmax layers, using info either in self.softmax_dimensions or by combining metadata info of labelname->numclasses and the outputclassnames
        """
        # NOTE, for re-id type models...multiclassification model will anyway yield the features with softmax outputs, so we don't have to worry about that...
        # For pure-reid model, probably best to use ClassificationResNet and modify to use no softmax...TODO this is a future step...
        tsoftmax = [None] * self.number_outputs
        for idx, fc_dimension in enumerate(self.softmax_dimensions):
            tsoftmax[idx] = nn.Linear(
                self.embedding_dimensions, fc_dimension, bias=False
            )
            tsoftmax[idx].apply(self.weights_init_softmax)
        self.softmax = nn.ModuleList(tsoftmax)

    def base_forward(self, x):
        features = self.gap(self.base(x))
        features = features.view(features.shape[0], -1)
        features = self.emb_linear(features)
        return features

    def forward_impl(self, x, **kwargs):
        features = self.base_forward(x)

        # if self.feat_norm is not None: <-- no need, identity
        features = self.feat_norm(features)

        softmax_logits = [None] * self.number_outputs
        for idx, softmaxlayer in enumerate(self.softmax):
            softmax_logits[idx] = softmaxlayer(features)
        return (
            softmax_logits,
            features,
            [],
        )  # soft logits are the softmax logits we will use to for training. We can use features to store the historical probability????



"""Contains code to build a resnet

  Typical usage example:

  model = resnet()
  x = torch.randn((3,100,100))
  features = model(x)
"""

from typing import List
from torch import nn
import torch

from ednaml.utils.blocks import InputAttention
from ednaml.utils.blocks import ResnetBasicBlock as BasicBlock
from ednaml.utils.blocks import ResnetBottleneck as Bottleneck


class resnet(nn.Module):
    def __init__(
        self,
        block: nn.Module = Bottleneck,
        layers: List[int] = [3, 4, 6, 3],
        last_stride: int = 2,
        zero_init_residual: bool = False,
        top_only: bool = True,
        num_classes: bool = 1000,
        groups: int = 1,
        width_per_group: int = 64,
        replace_stride_with_dilation: List[int] = None,
        norm_layer: nn.Module = None,
        attention: str = None,
        input_attention: bool = None,
        secondary_attention: int = None,
        ia_attention: bool = None,
        part_attention: bool = None,
        **kwargs
    ):
        """Initializes the resnet model.


        Args:
            block (nn.Module, optional): The block (BasicBlock or Bottleneck) used for the internal Resnet architecture. Defaults to ResnetBottleneck.
            layers (List[int], optional): Number of layers in each block. Defaults to [3, 4, 6, 3].
            last_stride (int, optional): The stride for the last block. Defaults to 2.
            zero_init_residual (bool, optional): Whether to initialize network with only zeros. Unused. Defaults to False.
            top_only (bool, optional): Whether to keep only the feature extractor block. Defaults to True.
            num_classes (bool, optional): Number of classes for the imagenet layer. Unused. Defaults to 1000.
            groups (int, optional): see nn.conv2D. Defaults to 1.
            width_per_group (int, optional): see nn.conv2D. Defaults to 64.
            replace_stride_with_dilation (List[int], optional): Whether to replace stride with dilation for each block. Defaults to None.
            norm_layer (nn.Module, optional): The default normalization layer. If None, uses batchnorm. Defaults to None.
            attention (str, optional): What attention to use, among `cbam`, `dbam`. Defaults to None.
            input_attention (bool, optional): Whether to use input attention at the first resnet block. Defaults to None.
            secondary_attention (int, optional): Whether to use secondary attention to apply `attention` to the specific resnet block only. Defaults to None.
            ia_attention (bool, optional): Whether to use input attention at the first conv layer. Exclusive with `input_attention`. Defaults to None.
            part_attention (bool, optional): Whether to use the local attention module. Defaults to None.

        Raises:
            ValueError: If attention blocks are not self-consistent. Specifically, the following rules:
                - cannot have both `ia_attention` and `input_attention`.
                - cannot have `part_attention` with `attention`, unless `secondary_attention`!=1
            ValueError: If `replace_stride_with_dilation` is not a 3-tuple or None.
        """
        super().__init__()

        self.attention = attention
        self.input_attention = input_attention
        self.secondary_attention = secondary_attention
        self.block = block
        self.inplanes = 64
        if norm_layer is None:
            self._norm_layer = nn.BatchNorm2d
        # elif norm_layer == "ln":
        #    self._norm_layer = nn.LayerNorm
        self.dilation = 1
        if replace_stride_with_dilation is None:
            replace_stride_with_dilation = [False, False, False]
        if len(replace_stride_with_dilation) != 3:
            raise ValueError(
                "replace_stride_with_dilation should be `None` or a 3-element tuple. Got {}".format(
                    replace_stride_with_dilation
                )
            )
        self.groups = groups
        self.base_width = width_per_group

        self.conv1 = nn.Conv2d(
            3, self.inplanes, kernel_size=7, stride=2, padding=3, bias=False
        )
        # if norm_layer == "gn":
        #    self.bn1 = nn.GroupNorm2d
        self.bn1 = nn.BatchNorm2d(self.inplanes)
        self.relu1 = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)

        self.ia_attention = ia_attention
        self.part_attention = part_attention

        # Make sure ia and input_attention do not conflict
        if self.ia_attention is not None and self.input_attention is not None:
            raise ValueError("Cannot have both ia_attention and input_attention.")
        if self.part_attention is not None and (
            self.attention is not None and self.secondary_attention is None
        ):
            raise ValueError("Cannot have part-attention with CBAM everywhere")
        if self.part_attention is not None and (
            self.attention is not None and self.secondary_attention == 1
        ):
            raise ValueError("Cannot have part-attention with CBAM-Early")

        # Create true IA
        if self.ia_attention:
            self.ia_attention = InputAttention(self.inplanes)  # 64, set above
        else:
            self.ia_attention = None

        att = self.attention
        if (
            secondary_attention is not None and secondary_attention != 1
        ):  # leave alone if sec attention not set
            att = None
        self.layer1 = self._make_layer(
            self.block,
            64,
            layers[0],
            attention=att,
            input_attention=self.input_attention,
            part_attention=self.part_attention,
        )
        att = self.attention
        if (
            secondary_attention is not None and secondary_attention != 2
        ):  # leave alone if sec attention not set
            att = None
        self.layer2 = self._make_layer(
            self.block,
            128,
            layers[1],
            stride=2,
            attention=att,
            dilate=replace_stride_with_dilation[0],
        )
        att = self.attention
        if (
            secondary_attention is not None and secondary_attention != 3
        ):  # leave alone if sec attention not set
            att = None
        self.layer3 = self._make_layer(
            self.block,
            256,
            layers[2],
            stride=2,
            attention=att,
            dilate=replace_stride_with_dilation[1],
        )
        att = self.attention
        if (
            secondary_attention is not None and secondary_attention != 4
        ):  # leave alone if sec attention not set
            att = None
        self.layer4 = self._make_layer(
            self.block,
            512,
            layers[3],
            stride=last_stride,
            attention=att,
            dilate=replace_stride_with_dilation[2],
        )

        self.top_only = top_only
        self.avgpool, self.fc = None, None

        if not self.top_only:
            self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
            self.fc = nn.Linear(512 * block.expansion, num_classes)

    def _make_layer(
        self,
        block: nn.Module,
        planes: int,
        blocks: int,
        stride: int = 1,
        dilate: bool = False,
        attention: str = None,
        input_attention: bool = False,
        ia_attention: bool = False,
        part_attention: bool = False,
    ) -> nn.Sequential:
        """Creates a resnet block

        Args:
            block (nn.Module): The block (BasicBlock or Bottleneck) used for the internal Resnet architecture. Defaults to ResnetBottleneck.
            planes (int): Number of input depth
            blocks (int): Number of blocks in this ResnetBlock
            stride (int, optional): Stride for the conv layers. Defaults to 1.
            dilate (bool, optional): Dilation for the conv layers. Defaults to False.
            attention (str, optional): Which of `cbam`, `dbam` attention to use. Defaults to None.
            input_attention (bool, optional): Whether to use `input_attention`. Defaults to False.
            ia_attention (bool, optional): Whether to use `ia_attention`. Unused. Defaults to False.
            part_attention (bool, optional): Whether to use local attention. Defaults to False.

        Returns:
            nn.Sequential: The layers comprising this Resnet Block as an nn.Sequential
        """
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(
                    self.inplanes,
                    planes * block.expansion,
                    kernel_size=1,
                    stride=stride,
                    bias=False,
                ),
                self._norm_layer(planes * block.expansion),
            )

        layers = []
        layers.append(
            block(
                self.inplanes,
                planes,
                stride,
                downsample,
                groups=self.groups,
                base_width=self.base_width,
                dilation=previous_dilation,
                norm_layer=self._norm_layer,
                attention=attention,
                input_attention=input_attention,
                part_attention=part_attention,
            )
        )
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(
                block(
                    self.inplanes,
                    planes,
                    groups=self.groups,
                    base_width=self.base_width,
                    dilation=self.dilation,
                    norm_layer=self._norm_layer,
                    attention=attention,
                )
            )
        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)

        if self.ia_attention is not None:
            x = self.ia_attention(x) * x
        x = self.bn1(x)
        x = self.relu1(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        if not self.top_only:
            x = self.avgpool(x)
            x = torch.flatten(x, 1)
            x = self.fc(x)
        return x

    def load_param(self, weights_path: str):
        """Loads parameters frm saved weights file

        Args:
            weights_path (str): Path to the weights file
        """
        param_dict = torch.load(weights_path)
        for i in param_dict:
            if "fc" in i and self.top_only:
                continue
            self.state_dict()[i].copy_(param_dict[i])


def _resnet(arch, block, layers, pretrained, progress, **kwargs) -> resnet:
    """Builds a resnet

    Args:
        arch (str): Architecture base. Unused.
        block (nn.Module): The class of the ResnetBlock (BasicBlock or Bottleneck)
        layers (List[int]): The layers for each ResnetBlock
        pretrained (bool): Unused
        progress (bool): Unused

    Returns:
        resnet: The `resnet` model
    """
    model = resnet(block, layers, **kwargs)
    return model


def resnet18(pretrained=False, progress=True, **kwargs):
    r"""ResNet-18 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _resnet("resnet18", BasicBlock, [2, 2, 2, 2], pretrained, progress, **kwargs)


def resnet34(pretrained=False, progress=True, **kwargs):
    r"""ResNet-34 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _resnet("resnet34", BasicBlock, [3, 4, 6, 3], pretrained, progress, **kwargs)


def resnet50(pretrained=False, progress=True, **kwargs):
    r"""ResNet-50 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _resnet("resnet50", Bottleneck, [3, 4, 6, 3], pretrained, progress, **kwargs)


def resnet101(pretrained=False, progress=True, **kwargs):
    r"""ResNet-101 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _resnet(
        "resnet101", Bottleneck, [3, 4, 23, 3], pretrained, progress, **kwargs
    )


def resnet152(pretrained=False, progress=True, **kwargs):
    r"""ResNet-152 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _resnet(
        "resnet152", Bottleneck, [3, 8, 36, 3], pretrained, progress, **kwargs
    )


def resnext50_32x4d(pretrained=False, progress=True, **kwargs):
    r"""ResNeXt-50 32x4d model from
    `"Aggregated Residual Transformation for Deep Neural Networks" <https://arxiv.org/pdf/1611.05431.pdf>`_
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    kwargs["groups"] = 32
    kwargs["width_per_group"] = 4
    return _resnet(
        "resnext50_32x4d", Bottleneck, [3, 4, 6, 3], pretrained, progress, **kwargs
    )


def resnext101_32x8d(pretrained=False, progress=True, **kwargs):
    r"""ResNeXt-101 32x8d model from
    `"Aggregated Residual Transformation for Deep Neural Networks" <https://arxiv.org/pdf/1611.05431.pdf>`_
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    kwargs["groups"] = 32
    kwargs["width_per_group"] = 8
    return _resnet(
        "resnext101_32x8d", Bottleneck, [3, 4, 23, 3], pretrained, progress, **kwargs
    )


def wide_resnet50_2(pretrained=False, progress=True, **kwargs):
    r"""Wide ResNet-50-2 model from
    `"Wide Residual Networks" <https://arxiv.org/pdf/1605.07146.pdf>`_
    The model is the same as ResNet except for the bottleneck number of channels
    which is twice larger in every block. The number of channels in outer 1x1
    convolutions is the same, e.g. last block in ResNet-50 has 2048-512-2048
    channels, and in Wide ResNet-50-2 has 2048-1024-2048.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    kwargs["width_per_group"] = 64 * 2
    return _resnet(
        "wide_resnet50_2", Bottleneck, [3, 4, 6, 3], pretrained, progress, **kwargs
    )


def wide_resnet101_2(pretrained=False, progress=True, **kwargs):
    r"""Wide ResNet-101-2 model from
    `"Wide Residual Networks" <https://arxiv.org/pdf/1605.07146.pdf>`_
    The model is the same as ResNet except for the bottleneck number of channels
    which is twice larger in every block. The number of channels in outer 1x1
    convolutions is the same, e.g. last block in ResNet-50 has 2048-512-2048
    channels, and in Wide ResNet-50-2 has 2048-1024-2048.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    kwargs["width_per_group"] = 64 * 2
    return _resnet(
        "wide_resnet101_2", Bottleneck, [3, 4, 23, 3], pretrained, progress, **kwargs
    )



from ednaml.crawlers import Crawler
from torch.utils.data import Dataset
from torch.utils.data import DataLoader

from ednaml.generators import ImageGenerator


class DataReader:
    name: str = "DataReader"
    CRAWLER: Crawler = Crawler
    DATASET: Dataset = Dataset
    DATALOADER: DataLoader = DataLoader
    GENERATOR: ImageGenerator = ImageGenerator

    def __init__(self):
        pass


"""
from .VehicleColor import VehicleColor
from .CompCars import CompCars
from .VehicleID import VehicleID
from .VeRi import VeRi
"""



import os.path as osp
from typing import List
import torch

# from torchvision.io import read_image
from PIL import Image
from PIL import ImageFile

ImageFile.LOAD_TRUNCATED_IMAGES = True
from torch.utils.data import Dataset as TorchDataset
from torch.utils.data import DataLoader as TorchDataLoader
from ednaml.utils.LabelMetadata import LabelMetadata
from ednaml.generators import ImageGenerator


class ClassificationDataset(TorchDataset):
    def __init__(self, dataset, transform=None, **kwargs):
        self.dataset = (
            dataset  # this is crawler.metadata["train"]["crawl"] -> [(), (), ()]
        )
        self.transform = transform
        self.pathidx = kwargs.get("pathidx", 0)
        self.annotationidx = kwargs.get("annotationidx", 1)

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx):
        # NOTE we will need to do our idx transforms here as well...
        # NOTE or leave it to be done in the crawler, and leave the generator as pristine and untouched???
        return (
            self.transform(self.load(self.dataset[idx][self.pathidx])),
            self.dataset[idx][self.annotationidx],
        )  # NOTE make this return a tuple for deploy (tuple of compressed), label

    def load(self, img):
        if not osp.exists(img):
            raise IOError("{img} does not exist in path".format(img=img))
        img_load = Image.open(img).convert("RGB")
        return img_load


class ClassificationGenerator(ImageGenerator):
    def __init__(
        self,
        gpus,
        i_shape=(208, 208),
        channels=3,
        normalization_mean=0.5,
        normalization_std=0.5,
        normalization_scale=1.0 / 255.0,
        h_flip=0.5,
        t_crop=True,
        rea=True,
        **kwargs
    ):
        """ Data generator for training and testing. Works with the <>. Should work with any crawler working on VeRi-like data. Not yet tested with VehicleID. Only  use with VeRi.

        Generates batches of batch size CONFIG.TRANSFORMATION.BATCH_SIZE, with CONFIG.TRANSFORMATION.INSTANCE unique ids. So if BATCH_SIZE=36 and INSTANCE=6, then generate batch of 36 images, with 6 identities, 6 image per identity. See arguments of setup function for INSTANCE.

        Args:
            gpus (int): Number of GPUs
            i_shape (int, int): 2D Image shape
            normalization_mean (float): Value to pass as mean normalization parameter to pytorch Normalization
            normalization_std (float): Value to pass as std normalization parameter to pytorch Normalization
            normalization_scale (float): Value to pass as scale normalization parameter. Not used.
            h_flip (float): Probability of horizontal flip for image
            t_crop (bool): Whether to include random cropping
            rea (bool): Whether to include random erasing augmentation (at 0.5 prob)
        
        """
        super().__init__(
            gpus,
            i_shape=i_shape,
            channels=channels,
            normalization_mean=normalization_mean,
            normalization_std=normalization_std,
            normalization_scale=normalization_scale,
            h_flip=h_flip,
            t_crop=t_crop,
            rea=rea,
            **kwargs
        )

    def buildDataset(
        self, datacrawler, mode: str, transform: List[object], **kwargs
    ) -> TorchDataset:
        if mode == "train":
            return ClassificationDataset(
                datacrawler.metadata[mode]["crawl"], transform, **kwargs
            )
        elif mode == "test":
            # For testing, we combine images in the query and testing set to generate batches
            return ClassificationDataset(
                datacrawler.metadata["val"]["crawl"]
                + datacrawler.metadata[mode]["crawl"],
                transform,
                **kwargs
            )
        elif mode == "full":
            return ClassificationDataset(
                datacrawler.metadata["val"]["crawl"]
                + datacrawler.metadata["train"]["crawl"]
                + datacrawler.metadata["test"]["crawl"],
                transform,
                **kwargs
            )
        else:
            raise NotImplementedError()

    def buildDataLoader(self, dataset, mode, batch_size, **kwargs):
        if mode == "train":
            return TorchDataLoader(
                dataset,
                batch_size=batch_size * self.gpus,
                shuffle=True,
                num_workers=self.workers,
                collate_fn=self.collate_simple,
            )
        elif mode == "test":
            return TorchDataLoader(
                dataset,
                batch_size=batch_size * self.gpus,
                shuffle=True,
                num_workers=self.workers,
                collate_fn=self.collate_simple,
            )
        elif mode == "full":
            return TorchDataLoader(
                dataset,
                batch_size=batch_size * self.gpus,
                shuffle=True,
                num_workers=self.workers,
                collate_fn=self.collate_simple,
            )
        else:
            raise NotImplementedError()

    def getNumEntities(self, datacrawler, mode, **kwargs):
        if mode in ["train", "test", "full"]:
            label_dict = {
                item: {"classes": datacrawler.metadata[mode]["classes"][item]}
                for item in [kwargs.get("classificationclass", "color")]
            }
            return LabelMetadata(label_dict=label_dict)
            # TODO fix this
        else:
            raise NotImplementedError()

    def collate_simple(self, batch):
        img, class_id, = zip(
            *batch
        )  # NOTE for deploy, this is a tuple, label, i.e. (img, img, img), class
        class_id = torch.tensor(class_id, dtype=torch.int64)
        return (
            torch.stack(img, dim=0),
            class_id,
        )  # NOTE so this has to be modified...for tuple members...



from ednaml.trainer.BaseTrainer import BaseTrainer

from ednaml.trainer.ClassificationTrainer import ClassificationTrainer
from ednaml.trainer.MultiClassificationTrainer import MultiClassificationTrainer
from ednaml.trainer.MultiBranchTrainer import MultiBranchTrainer



from torch import nn
import torch
from ednaml.loss import Loss


class SoftmaxLabelSmooth(Loss):
    """Softmax with label smoothing

    Performs softmax with label smoothing.

    Args (kwargs only):
        softmax_dimensions (int): Dimension of the softmax layer. Used for the smoothing constant scaling.
        eps (float): Smothing constant. Default 0.1

    Methods: 
        __call__: Returns loss given logits and labels.

    """

    def __init__(self, lossname, metadata, **kwargs):
        super().__init__()
        self.softmax_dimensions = kwargs.get("softmax_dimensions", None)
        if self.softmax_dimensions is None:
            self.softmax_dimensions = metadata.getLabelDimensions(lossname)
        self.eps = kwargs.get("eps", 0.1)
        self.logsoftmax = nn.LogSoftmax(dim=1)

    def forward(self, logits, labels):
        """
        Args:
            logits: prediction matrix (before softmax) with shape (batch_size, softmax_dimensions)
            labels: ground truth labels with shape (batch_size)
        """
        llogits = logits[labels >= 0]
        llabels = labels[labels >= 0]
        log_probs = self.logsoftmax(llogits)
        llabels = torch.zeros(log_probs.size()).scatter_(
            1, llabels.unsqueeze(1).data.cpu(), 1
        )
        llabels = llabels.cuda()
        llabels = (1 - self.eps) * llabels + self.eps / self.softmax_dimensions
        loss = (-llabels * log_probs).mean(0).sum()
        return loss



from typing import Dict


class ExecutionDatareaderConfig:
    DATAREADER: str
    CRAWLER_ARGS: Dict[str, str]
    DATASET_ARGS: Dict[str, str]
    GENERATOR: str
    GENERATOR_ARGS: Dict[str, str]
    DATALOADER_ARGS: Dict[str, str]

    def __init__(self, datareader_dict):
        self.DATAREADER = datareader_dict.get("DATAREADER")
        self.CRAWLER_ARGS = datareader_dict.get("CRAWLER_ARGS", {})
        self.DATASET_ARGS = datareader_dict.get("DATASET_ARGS", {})
        self.GENERATOR = datareader_dict.get("GENERATOR", "ClassificationGenerator")
        self.GENERATOR_ARGS = datareader_dict.get("GENERATOR_ARGS", {})
        self.DATALOADER_ARGS = datareader_dict.get("DATALOADER_ARGS", {})



from ednaml.optimizer import BaseOptimizer


class ClassificationOptimizer(BaseOptimizer):
    """ Optimizer Builder for CoLabel experiments.

    """

    def __init__(
        self,
        name,
        optimizer,
        base_lr,
        lr_bias,
        gpus,
        weight_decay,
        weight_bias,
        opt_kwargs,
    ):
        """ Initializes the optimizer builder.

        Args:
            base_lr (float): Base learning rate for optimizer
            lr_bias (float): Multiplicative factor for bias parameters
            gpus (int): Number of GPUs for lr scaling
            weight_decay (float): Weight decay for decoupled weight decay optimizers like AdamW
            weight_bias (float): Multiplicative factor for bias parameters in weight decay optimizers

        Methods:
            build:  builds an optimizer given optimizer name and torch model

        """
        super().__init__(
            name,
            optimizer,
            base_lr,
            lr_bias,
            gpus,
            weight_decay,
            weight_bias,
            opt_kwargs,
        )



import os
from typing import Tuple
from ednaml.config.EdnaMLConfig import EdnaMLConfig


class SaveMetadata:
    """Contains paths for saving EdnaML models, loggers, metadata, and training information
    """

    MODEL_VERSION: str
    MODEL_CORE_NAME: str
    MODEL_BACKBONE: str
    MODEL_QUALIFIER: str
    DRIVE_BACKUP: bool

    MODEL_SAVE_NAME: str
    MODEL_SAVE_FOLDER: str
    LOGGER_SAVE_NAME: str
    CHECKPOINT_DIRECTORY: str

    def __init__(self, cfg: EdnaMLConfig):
        (
            self.MODEL_SAVE_NAME,
            self.MODEL_SAVE_FOLDER,
            self.LOGGER_SAVE_NAME,
            self.CHECKPOINT_DIRECTORY,
        ) = SaveMetadata.generate_save_names_from_config(cfg)

        self.MODEL_VERSION = cfg.SAVE.MODEL_VERSION
        self.MODEL_CORE_NAME = cfg.SAVE.MODEL_CORE_NAME
        self.MODEL_BACKBONE = cfg.SAVE.MODEL_BACKBONE
        self.MODEL_QUALIFIER = cfg.SAVE.MODEL_QUALIFIER
        self.DRIVE_BACKUP = cfg.SAVE.DRIVE_BACKUP

    @staticmethod
    def generate_save_names_from_config(cfg: EdnaMLConfig) -> Tuple[str]:

        MODEL_SAVE_NAME = "%s-v%i" % (cfg.SAVE.MODEL_CORE_NAME, cfg.SAVE.MODEL_VERSION)
        MODEL_SAVE_FOLDER = "%s-v%i-%s-%s" % (
            cfg.SAVE.MODEL_CORE_NAME,
            cfg.SAVE.MODEL_VERSION,
            cfg.SAVE.MODEL_BACKBONE,
            cfg.SAVE.MODEL_QUALIFIER,
        )
        LOGGER_SAVE_NAME = "%s-v%i-%s-%s-logger.log" % (
            cfg.SAVE.MODEL_CORE_NAME,
            cfg.SAVE.MODEL_VERSION,
            cfg.SAVE.MODEL_BACKBONE,
            cfg.SAVE.MODEL_QUALIFIER,
        )
        if cfg.SAVE.DRIVE_BACKUP:
            CHECKPOINT_DIRECTORY = os.path.join(
                cfg.SAVE.CHECKPOINT_DIRECTORY, MODEL_SAVE_FOLDER
            )
        else:
            CHECKPOINT_DIRECTORY = ""
        return (
            MODEL_SAVE_NAME,
            MODEL_SAVE_FOLDER,
            LOGGER_SAVE_NAME,
            CHECKPOINT_DIRECTORY,
        )



from torch import nn


class ChannelAttention(nn.Module):
    """Channel Attention module that forms part of CBAM attention.
    See the ECCV paper at https://openaccess.thecvf.com/content_ECCV_2018/papers/Sanghyun_Woo_Convolutional_Block_Attention_ECCV_2018_paper.pdf
    """

    def __init__(self, in_planes, ratio=16):
        super(ChannelAttention, self).__init__()
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.max_pool = nn.AdaptiveMaxPool2d(1)

        self.fc1 = nn.Conv2d(in_planes, in_planes // ratio, 1, bias=False)
        self.relu1 = nn.ReLU()
        self.fc2 = nn.Conv2d(in_planes // ratio, in_planes, 1, bias=False)

        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        avg_out = self.fc2(self.relu1(self.fc1(self.avg_pool(x))))
        max_out = self.fc2(self.relu1(self.fc1(self.max_pool(x))))
        out = avg_out + max_out
        return self.sigmoid(out)



from typing import List

from ednaml.config.ConfigDefaults import ConfigDefaults


class TransformationConfig:
    SHAPE: List[int]
    NORMALIZATION_MEAN: List[int]
    NORMALIZATION_STD: List[int]
    NORMALIZATION_SCALE: int
    H_FLIP: float
    T_CROP: bool
    RANDOM_ERASE: bool
    RANDOM_ERASE_VALUE: float
    CHANNELS: int
    BATCH_SIZE: int
    WORKERS: int

    def __init__(self, transformation_dict, defaults: ConfigDefaults):
        self.SHAPE = transformation_dict.get("SHAPE", defaults.SHAPE)
        self.NORMALIZATION_MEAN = transformation_dict.get(
            "NORMALIZATION_MEAN", defaults.NORMALIZATION_MEAN
        )
        self.NORMALIZATION_STD = transformation_dict.get(
            "NORMALIZATION_STD", defaults.NORMALIZATION_STD
        )
        self.NORMALIZATION_SCALE = transformation_dict.get(
            "NORMALIZATION_SCALE", defaults.NORMALIZATION_SCALE
        )
        self.H_FLIP = transformation_dict.get("H_FLIP", defaults.H_FLIP)
        self.T_CROP = transformation_dict.get("T_CROP", defaults.T_CROP)
        self.RANDOM_ERASE = transformation_dict.get(
            "RANDOM_ERASE", defaults.RANDOM_ERASE
        )
        self.RANDOM_ERASE_VALUE = transformation_dict.get(
            "RANDOM_ERASE_VALUE", defaults.RANDOM_ERASE_VALUE
        )
        self.CHANNELS = transformation_dict.get("CHANNELS", defaults.CHANNELS)
        self.BATCH_SIZE = transformation_dict.get("BATCH_SIZE", defaults.BATCH_SIZE)
        self.WORKERS = transformation_dict.get("WORKERS", defaults.WORKERS)



from sched import scheduler
from typing import Dict

from ednaml.config.ConfigDefaults import ConfigDefaults


class SchedulerConfig:
    SCHEDULER_NAME: str
    LR_SCHEDULER: str
    LR_KWARGS: Dict[str, str]

    def __init__(self, scheduler_dict, defaults: ConfigDefaults):
        self.SCHEDULER_NAME = scheduler_dict.get(
            "SCHEDULER_NAME", defaults.SCHEDULER_NAME
        )
        self.LR_SCHEDULER = scheduler_dict.get("LR_SCHEDULER", defaults.LR_SCHEDULER)
        self.LR_KWARGS = scheduler_dict.get("LR_KWARGS", defaults.LR_KWARGS)



import glob
import math
import os
import random
import ednaml.utils.splits.cub200


class CUB200_2011DataCrawler:
    def __init__(self, data_folder="CUB_200_2011", **kwargs):
        self.metadata = {}

        self.data_folder = data_folder
        self.image_folder = os.path.join(self.data_folder, "images")

        self.logger = kwargs.get("logger")

        self.__verify(self.data_folder)
        self.__verify(self.image_folder)

        self.crawl()

    def __verify(self, folder):
        if not os.path.exists(folder):
            raise IOError(
                "Folder {data_folder} does not exist".format(data_folder=folder)
            )
        else:
            self.logger.info("Found {data_folder}".format(data_folder=folder))

    def crawl(self,):
        self.metadata["train"], self.metadata["test"], self.metadata["query"] = (
            {},
            {},
            {},
        )
        self.__crawl(self.image_folder)

        self.logger.info(
            "Train\tPIDS: {:6d}\tCIDS: {:6d}\tIMGS: {:8d}".format(
                self.metadata["train"]["pids"],
                self.metadata["train"]["cids"],
                self.metadata["train"]["imgs"],
            )
        )
        self.logger.info(
            "Test \tPIDS: {:6d}\tCIDS: {:6d}\tIMGS: {:8d}".format(
                self.metadata["test"]["pids"],
                self.metadata["test"]["cids"],
                self.metadata["test"]["imgs"],
            )
        )
        self.logger.info(
            "Query\tPIDS: {:6d}\tCIDS: {:6d}\tIMGS: {:8d}".format(
                self.metadata["query"]["pids"],
                self.metadata["query"]["cids"],
                self.metadata["query"]["imgs"],
            )
        )

    def __crawl(self, image_folder):
        # Data/CUB_200_2011/images contains one folder for each of the 200 classes...
        crawler = []
        class_list = glob.glob(image_folder + "/*")
        for class_folder in class_list:
            folder_name = os.path.basename(class_folder)
            class_name = int(folder_name.split(".")[0])
            image_list = glob.glob(class_folder + "/*.jpg")
            crawler += [(item, class_name, 0) for item in image_list]

        # crawler is a list of 3-tuples. Length of N=number of images.
        # Each tuple is (path/to/image, PID, CID)   PID --> class (from 0-200), CID --> 0
        # CID is a holdover from other crawlers used for re-id task, where cid, or camera-id is required. To maintain compatibility with SequencedGenerator (which expects CID) until I write a generator for Cars196
        # PID is similarity from person-reid, where PID stands for person ID. In this case, it is a unique class
        trainvalsplits = {
            int(item.split(".")[0]): idx
            for idx, item in enumerate(ednaml.utils.splits.cub200.trainval)
        }
        testsplits = {
            int(item.split(".")[0]): idx + len(trainvalsplits)
            for idx, item in enumerate(ednaml.utils.splits.cub200.query)
        }
        train_crawler = [
            (item[0], trainvalsplits[item[1]], item[2])
            for item in crawler
            if item[1] in trainvalsplits
        ]
        test_crawler = [
            (item[0], testsplits[item[1]], item[2])
            for item in crawler
            if item[1] in testsplits
        ]

        random.shuffle(train_crawler)
        split = 0.7
        split_idx = math.ceil(split * len(train_crawler))

        self.metadata["test"]["crawl"] = train_crawler[split_idx:]
        self.metadata["test"]["pids"] = len(trainvalsplits)
        self.metadata["test"]["cids"] = 1
        self.metadata["test"]["imgs"] = len(self.metadata["test"]["crawl"])
        train_crawler = train_crawler[:split_idx]

        self.metadata["query"]["crawl"] = test_crawler
        self.metadata["query"]["pids"] = len(testsplits)
        self.metadata["query"]["cids"] = 1
        self.metadata["query"]["imgs"] = len(self.metadata["query"]["crawl"])

        self.metadata["train"]["crawl"] = train_crawler[:split_idx]
        self.metadata["train"]["pids"] = len(trainvalsplits)
        self.metadata["train"]["cids"] = 1
        self.metadata["train"]["imgs"] = len(train_crawler[:split_idx])



import glob
import math
import os
import random

import ednaml.utils.splits.sun


class SUNDataCrawler:
    def __init__(self, data_folder="SUNAttributeDB_Images", **kwargs):
        self.metadata = {}

        self.data_folder = data_folder
        self.image_folder = os.path.join(self.data_folder, "images")

        self.logger = kwargs.get("logger")

        self.__verify(self.data_folder)
        self.__verify(self.image_folder)

        self.crawl()

    def __verify(self, folder):
        if not os.path.exists(folder):
            raise IOError(
                "Folder {data_folder} does not exist".format(data_folder=folder)
            )
        else:
            self.logger.info("Found {data_folder}".format(data_folder=folder))

    def crawl(self,):
        self.metadata["train"], self.metadata["test"], self.metadata["query"] = (
            {},
            {},
            {},
        )
        self.__crawl(self.image_folder)

        self.logger.info(
            "Train\tPIDS: {:6d}\tCIDS: {:6d}\tIMGS: {:8d}".format(
                self.metadata["train"]["pids"],
                self.metadata["train"]["cids"],
                self.metadata["train"]["imgs"],
            )
        )
        self.logger.info(
            "Test \tPIDS: {:6d}\tCIDS: {:6d}\tIMGS: {:8d}".format(
                self.metadata["test"]["pids"],
                self.metadata["test"]["cids"],
                self.metadata["test"]["imgs"],
            )
        )
        self.logger.info(
            "Query\tPIDS: {:6d}\tCIDS: {:6d}\tIMGS: {:8d}".format(
                self.metadata["query"]["pids"],
                self.metadata["query"]["cids"],
                self.metadata["query"]["imgs"],
            )
        )

    def __crawl(self, image_folder):
        # Data/CUB_200_2011/images contains folders alphabetically...
        traincrawler = []
        querycrawler = []

        for idx, class_name in enumerate(ednaml.utils.splits.sun.trainval):
            # since the proposed split names are not separated by folders, we have to manually extract folders...
            alphabetical = class_name[0]
            directory_splits = class_name.split("_")
            path_proposal = self.get_true_path(
                alphabetical, directory_splits, image_folder
            )
            if path_proposal == False:
                print(alphabetical, directory_splits, image_folder)
                raise ValueError()
            image_list = glob.glob(path_proposal + "/*.jpg")
            traincrawler += [(item, idx, 0) for item in image_list]

        for idx, class_name in enumerate(ednaml.utils.splits.sun.query):
            # since the proposed split names are not separated by folders, we have to manually extract folders...
            alphabetical = class_name[0]
            directory_splits = class_name.split("_")
            path_proposal = self.get_true_path(
                alphabetical, directory_splits, image_folder
            )
            if path_proposal == False:
                raise ValueError()
            image_list = glob.glob(path_proposal + "/*.jpg")
            querycrawler += [
                (item, idx + len(ednaml.utils.splits.sun.trainval), 0)
                for item in image_list
            ]

        random.shuffle(traincrawler)
        split = 0.9
        split_idx = math.ceil(split * len(traincrawler))

        self.metadata["test"]["crawl"] = traincrawler[split_idx:]
        self.metadata["test"]["pids"] = len(ednaml.utils.splits.sun.trainval)
        self.metadata["test"]["cids"] = 1
        self.metadata["test"]["imgs"] = len(self.metadata["test"]["crawl"])

        self.metadata["query"]["crawl"] = querycrawler
        self.metadata["query"]["pids"] = len(ednaml.utils.splits.sun.query)
        self.metadata["query"]["cids"] = 1
        self.metadata["query"]["imgs"] = len(self.metadata["query"]["crawl"])

        self.metadata["train"]["crawl"] = traincrawler[:split_idx]
        self.metadata["train"]["pids"] = len(ednaml.utils.splits.sun.trainval)
        self.metadata["train"]["cids"] = 1
        self.metadata["train"]["imgs"] = len(traincrawler[:split_idx])

    def get_true_path(self, alphabetical, directory_splits, image_folder):
        full_name = "_".join(directory_splits)
        path_proposal = os.path.join(image_folder, alphabetical)
        path_proposal = os.path.join(path_proposal, full_name)
        if os.path.exists(path_proposal):
            return path_proposal
        # So the full name doesn't work. We'll try to build it piece by piece...
        path_proposal = os.path.join(image_folder, alphabetical)
        directory_proposal = ""
        for idx, directory in enumerate(directory_splits):
            # Build the path with the current proposal
            if directory_proposal == "":
                directory_proposal = "_".join([] + [directory])
            else:
                directory_proposal = "_".join([directory_proposal] + [directory])
            proposal = os.path.join(path_proposal, directory_proposal)
            if os.path.exists(proposal):
                path_proposal = proposal
                # reset directory proposal
                directory_proposal = ""
            # If path does not exist, we try by appending the next directory split to the directory proposal...
        if not os.path.exists(path_proposal):
            return False
        return path_proposal



from ednaml.datareaders import DataReader
from crawlers import VehicleIDDataCrawler
from generators.ClassificationGenerator import (
    ClassificationDataset,
    ClassificationGenerator,
)


class VehicleID(DataReader):
    name: str = "VehicleID"
    CRAWLER = VehicleIDDataCrawler
    DATASET = ClassificationDataset
    GENERATOR = ClassificationGenerator

    def __init__(self):
        super().__init__()



from ednaml.datareaders import DataReader
from crawlers import VeRiDataCrawler
from generators.ClassificationGenerator import (
    ClassificationDataset,
    ClassificationGenerator,
)


class VeRi(DataReader):
    name: str = "VeRi"
    CRAWLER = VeRiDataCrawler
    DATASET = ClassificationDataset
    GENERATOR = ClassificationGenerator

    def __init__(self):
        super().__init__()



from typing import List
from torch import nn
from ednaml.backbones.multibranchresnet import multibranchresnet
from ednaml.models import ModelAbstract
from ednaml.utils import layers
import torch


class MultiBranchResnet(ModelAbstract):
    """Multibranch Resnet model, that performs multiple classifications with different branches. Branches may be fused as well.

    A MultiBranchResnet model is a base ResNet with multiple FC classification layers.

    Args: (TODO)
        base (str): The architecture base for resnet, i.e. resnet50, resnet18
        weights (str, None): Path to weights file for the architecture base ONLY. If not provided, base initialized with random values.
        normalization (str, None): Can be None, where it is torch's normalization. Else create a normalization layer. Supports: ["bn", "l2", "in", "gn", "ln"]
        metadata (Dict[str:int], None): FC dimensions for each classification, keyed by class names

    Kwargs (MODEL_KWARGS):
        - `number_branches`: This is the number of branches for the model
        - `branches`: A list, each element is the i-th branch's metadata
            - `name`: This is the name of this branch. This is used by model_builder to keep track of output names and labels. 
            - `number_outputs`: This is the number of classification outputs for this branch
            - `outputs`: A list, each element is the j-th output's metadata for this branch
                - `dimensions`: This is the number of classes for this output. This can be left blank if you want EdnaML to infer the size from the `LABEL` parameter below. 
                - `name`: This is the name of this output. 
                - `label`: This is the name of the label this output is tracking. <span style="color:magenta; font-weight:bold">THIS SHOULD CORRESPOND EXACTLY WITH `DATASET_ARGS.classificationclass`</span> labels.
        - `fuse`: **Bool**. Whether branch outputs are going to be fused
        - `fuse_outputs`: List of output names (not branch names) that are fused
        - `fuse_dimensions`: The dimensions of the fused output. If left blank, EdnaML will infer from `fuse_label`
        - `fuse_label`: The label that tracks the fused output 
        
        last_stride (int, 1): The final stride parameter for the architecture core. Should be one of 1 or 2.
        attention (str, None): The attention module to use. Only supports ['cbam', 'dbam']
        input_attention (bool, false): Whether to include the IA module
        secondary_attention (List[int], None): Whether to modify CBAM to apply it to specific Resnet basic blocks. None means CBAM is applied to all. Otherwise, CBAM is applied only to the basic blocks provided here in List.
        part_attention (bool): Whether to use local attention

    Default Kwargs (DO NOT CHANGE OR ADD TO MODEL_KWARGS):
        zero_init_residual (bool, false): Whether the final layer uses zero initialization
        top_only (bool, true): Whether to keep only the architecture base without imagenet fully-connected layers (1000 classes)
        num_classes (int, 1000): Number of features in final imagenet FC layer
        groups (int, 1): Used during resnet variants construction
        width_per_group (int, 64): Used during resnet variants construction
        replace_stride_with_dilation (bool, None): Well, replace stride with dilation...
        norm_layer (nn.Module, None): The normalization layer within resnet. Internally defaults to nn.BatchNorm2D

    Methods: 
        forward: Process a batch

    """

    model_name = "MultiBranchResNet"
    model_arch = "MultiBranchResNet"
    number_outputs = 1
    output_names = ["out0"]
    softmax_dimensions = [None]
    output_labels = ["color"]
    secondary_outputs = []
    base: multibranchresnet
    _internal_name_count = 0

    model_labelorder: List[str]

    def __init__(
        self, base="resnet50", weights=None, normalization=None, metadata=None, **kwargs
    ):
        """We will inherit the base construction from ClassificationResNet, and modify the softmax head.

        Args:
            base (str, optional): _description_. Defaults to 'resnet50'.
            weights (_type_, optional): _description_. Defaults to None.
            normalization (_type_, optional): _description_. Defaults to None.
            metadata (_type_, optional): _description_. Defaults to None.
        """

        super().__init__(
            base=base,
            weights=weights,
            normalization=normalization,
            metadata=metadata,
            **kwargs
        )

    def model_attributes_setup(self, **kwargs):

        self.embedding_dimensions = kwargs.get("embedding_dimensions", None)
        if self.normalization == "":
            self.normalization = None

        # So, things we need
        # num branches
        #
        branches = kwargs.get("branches")
        self.number_branches = kwargs.get("number_branches", 3)
        self.branches_meta = {branch["name"]: branch for branch in branches}
        self.branch_name_order = [branch["name"] for branch in branches]

        self.number_outputs = 0
        self.output_name_order = (
            []
        )  # order of outputs, by their name. We can look up specific metadata in outputs_meta,, such as label it is tracking
        self.outputs_meta = {}
        for branch_name in self.branch_name_order:
            nouts = self.branches_meta[branch_name]["number_outputs"]
            self.number_outputs += nouts
            out_dict = {
                out["name"]: out for out in self.branches_meta[branch_name]["outputs"]
            }
            out_order = [
                out["name"] for out in self.branches_meta[branch_name]["outputs"]
            ]
            self.output_name_order += out_order

            self.outputs_meta = dict(self.outputs_meta, **out_dict)
            self.branches_meta[branch_name]["output_nameorder"] = [
                item for item in out_order
            ]  # for list copy, instead of reference...this is the outputs for this branch
        # Now we adjust the dimensions, i.e. fix them if they do not exist
        for output in self.outputs_meta:
            if "dimensions" in self.outputs_meta[output]:
                if self.outputs_meta[output]["dimensions"] is None:
                    self.outputs_meta[output][
                        "dimensions"
                    ] = self.metadata.getLabelDimensions(
                        self.outputs_meta[output]["label"]
                    )
            else:
                self.outputs_meta[output][
                    "dimensions"
                ] = self.metadata.getLabelDimensions(self.outputs_meta[output]["label"])

        self.output_dimensions = []
        self.output_label_order = []
        for oname in self.output_name_order:
            self.output_dimensions += [self.outputs_meta[oname]["dimensions"]]
            self.output_label_order += [self.outputs_meta[oname]["label"]]

        # Now we have self.output_name_order, which contains output-names in order of how they will be output
        # Now we need to add the fused information
        self.branch_fuse = kwargs.get("fuse", False)
        self.branch_fuse_names = {item: 1 for item in kwargs.get("fuse_outputs", [])}
        self.branch_fuse_idx = []
        if self.branch_fuse:
            self.branch_fuse_idx = [
                idx
                for idx, b_name in enumerate(self.branch_name_order)
                if b_name in self.branch_fuse_names
            ]
        self.fuse_dimensions = kwargs.get("fuse_dimensions", None)
        self.fuse_label = kwargs.get("fuse_label", None)
        if self.fuse_dimensions is None:
            self.fuse_dimensions = self.metadata.getLabelDimensions(self.fuse_label)

        # need metadata for model_labelorder for the output
        self.model_labelorder = [item for item in self.output_label_order]
        self.model_nameorder = [item for item in self.output_name_order]
        if self.branch_fuse:
            self.model_labelorder += [self.fuse_label]
            self.model_nameorder += [kwargs.get("fuse_name", "fuse")]

        self.base = None
        self.branches = {}
        for bname in self.branch_name_order:
            self.branches[bname] = {}
            self.branches[bname]["gap"] = None
            self.branches[bname]["emb_linear"] = None
            self.branches[bname]["feat_norm"] = None
            self.branches[bname]["softmax"] = [None] * self.branches_meta[bname][
                "number_outputs"
            ]

        self.output_count = self.number_outputs
        self.feature_count = self.number_branches
        if self.branch_fuse:
            self.output_count += 1
            self.feature_count += 1

    def model_setup(self, **kwargs):
        self.build_base(**kwargs)
        self.build_normalization()
        self.build_softmax()
        self.build_fused()

    def build_base(self, **kwargs):
        """Build the model base.

        Builds the architecture base/core.
        """
        _resnet = __import__("backbones.multibranchresnet", fromlist=["resnet"])
        _resnet = getattr(_resnet, self.model_base)
        # Set up the resnet backbone
        self.base = _resnet(last_stride=1, **kwargs)
        if self.weights is not None:
            self.base.load_param(self.weights)

        if self.embedding_dimensions is None:
            self.embedding_dimensions = 512 * self.base.block.expansion
        self.fused_feat_dimensions = (
            len(self.branch_fuse_idx) * 512 * self.base.block.expansion
        )

        for bname in self.branch_name_order:
            self.branches[bname]["gap"] = nn.AdaptiveAvgPool2d(1)
            self.branches[bname]["emb_linear"] = torch.nn.Identity()

            if self.embedding_dimensions != 512 * self.base.block.expansion:
                self.branches[bname]["emb_linear"] = nn.Linear(
                    self.base.block.expansion * 512,
                    self.embedding_dimensions,
                    bias=False,
                )

            self.branches[bname]["feat_norm"] = None
            self.branches[bname]["softmax"] = [None] * self.branches_meta[bname][
                "number_outputs"
            ]

    def build_normalization(self):

        norm_func = nn.Module
        norm_args = {}
        norm_div = 1
        if self.normalization == "bn":
            norm_func = nn.BatchNorm1d
            norm_args = {"affine": True}
        elif self.normalization == "in":
            norm_func = layers.FixedInstanceNorm1d
            norm_args = {"affine": True}
        elif self.normalization == "gn":
            norm_div = 16
            norm_func = nn.GroupNorm
            norm_args = {"num_channels": self.embedding_dimensions, "affine": True}
        elif self.normalization == "ln":
            norm_func = nn.LayerNorm
            norm_args = {"elementwise_affine": True}
        elif self.normalization == "l2":
            norm_func = layers.L2Norm
            norm_args = {"scale": 1.0}
        elif self.normalization is None or self.normalization == "":
            norm_func = torch.nn.Identity
            norm_args = {}
        else:
            raise NotImplementedError()

        self.fused_feat_norm = norm_func(
            self.fused_feat_dimensions // norm_div, **norm_args
        )
        for bname in self.branch_name_order:
            self.branches[bname]["feat_norm"] = norm_func(
                self.embedding_dimensions // norm_div, **norm_args
            )

    def build_softmax(self):
        # We will build the softmax layers...
        # For this, we will go into the self.branches, and add the softmax there...
        for bname in self.branch_name_order:
            for idx, output_name in enumerate(
                self.branches_meta[bname]["output_nameorder"]
            ):
                self.branches[bname]["softmax"][idx] = nn.Linear(
                    self.embedding_dimensions,
                    self.outputs_meta[output_name]["dimensions"],
                    bias=False,
                )
                self.branches[bname]["softmax"][idx].apply(self.weights_init_softmax)
            self.branches[bname]["softmax"] = nn.ModuleList(
                self.branches[bname]["softmax"]
            )

    def build_fused(self):
        for bname in self.branch_name_order:
            self.branches[bname] = nn.ModuleDict(self.branches[bname])

        self.branches = nn.ModuleDict(self.branches)
        if self.branch_fuse:
            self.fused_feat_dimensions = (
                len(self.branch_fuse_idx) * 512 * self.base.block.expansion
            )
            self.softmax_fused = nn.Linear(
                self.fused_feat_dimensions, self.fuse_dimensions, bias=False
            )
            self.softmax_fused.apply(self.weights_init_softmax)

    def _internal_name_counter(self):
        out = "out" + str(self._internal_name_count)
        self._internal_name_count += 1
        return out

    def base_forward(self, x):
        features = self.base(x)  # features are a list

        for idx, bname in enumerate(self.branch_name_order):
            features[idx] = self.branches[bname]["gap"](features[idx])
            features[idx] = features[idx].view(features[idx].shape[0], -1)
            features[idx] = self.branches[bname]["emb_linear"](features[idx])

        return features

    def forward_impl(self, x, **kwargs):

        features = self.base_forward(x)
        fused_features = []
        if self.branch_fuse:
            fused_features += [
                self.fused_feat_norm(
                    torch.cat([features[idx] for idx in self.branch_fuse_idx], dim=1)
                )
            ]

        outputs = [None] * self.number_outputs
        out_idx = 0
        for b_idx, bname in enumerate(self.branch_name_order):
            features[b_idx] = self.branches[bname]["feat_norm"](features[b_idx])
            for o_idx, _ in enumerate(self.branches_meta[bname]["output_nameorder"]):
                outputs[out_idx] = self.branches[bname]["softmax"][o_idx](
                    features[b_idx]
                )
                out_idx += 1

        fused_outs = []
        if self.branch_fuse:
            fused_outs += [self.softmax_fused(fused_features[0])]

        return features + fused_features, outputs + fused_outs, []



from ednaml.config.ConfigDefaults import ConfigDefaults


class LoggingConfig:
    STEP_VERBOSE: int

    def __init__(self, logging_dict, defaults: ConfigDefaults):
        self.STEP_VERBOSE = logging_dict.get("STEP_VERBOSE", defaults.STEP_VERBOSE)
